# Copyright 2019-2021 Wingify Software Pvt. Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Module integrated in the vwo sdk for retrieving settings_file from the server
"""

from __future__ import print_function
import sys
import requests
from ..constants import constants
from ..helpers import generic_util, validate_util


def get(account_id, sdk_key):
    """ Get method to retrieve settings_file for customer from our server

    Args:
        account_id (string): Account ID of user
        sdk_key (string): Unique sdk key for user,
            can be retrieved from our webside

    Returns:
        json_string|None: Json representation of settings_file,
            as received from the website,
            None if no settings_file is found or sdk_key is incorrect
    """

    is_valid_account_id = validate_util.is_valid_number(account_id) or validate_util.is_valid_string(account_id)

    if not is_valid_account_id or not validate_util.is_valid_string(sdk_key):
        print(("account_id and sdk_key are required", "for fetching account settings. Aborting!"), file=sys.stderr)
        return "{}"

    protocol = constants.HTTPS_PROTOCOL
    hostname = constants.ENDPOINTS.BASE_URL
    path = constants.ENDPOINTS.ACCOUNT_SETTINGS

    parameters = {
        "a": account_id,
        "i": sdk_key,
        "r": generic_util.get_random_number(),
        "platform": constants.PLATFORM,
        "api-version": constants.API_VERSION,
    }
    server_url = protocol + hostname + path
    try:
        settings_file_response = requests.get(server_url, params=parameters)
        if settings_file_response.status_code != 200:
            print(
                "Request failed for fetching account settings. "
                "Got Status Code: {status_code} "
                "and message: {settings_file_response}.".format(
                    status_code=settings_file_response.status_code, settings_file_response=settings_file_response.text
                ),
                file=sys.stderr,
            )
        settings_file = settings_file_response.text
    except requests.exceptions.RequestException as e:
        print("Error fetching Settings File", e, file=sys.stderr)
        return "{}"
    return settings_file



# Copyright 2019-2021 Wingify Software Pvt. Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
from ..enums.log_message_enum import LogMessageEnum
from ..enums.file_name_enum import FileNameEnum
import logging


_VWO_LOG_FORMAT = "VWO-SDK - [%(levelname)s]: %(asctime)s %(message)s"
API_NAME = "SDK"
_DEFAULT_LOGGING_LEVEL = logging.ERROR
FILE = FileNameEnum.Logger.LoggerManager
_logger = None


def clearExistingLoggerInstance():
    """ Clears the existing logger """
    global _logger
    _logger = None


def getInstance(**kwargs):
    """ Always use this method to get an instance of VWOLogger. Pass a valid log_level
    or logger object when calling for first time to instantiate a VWOLogger object.
    If logger is passed, log_level is neglected.
    It throws exception when called with kwargs 2nd time.

    Keyword Args:
        log_level (vwo.LOG_LEVELS): A valid log level accepted by this SDK
        logger (object): An object capable of logging events

    Returns:
        VWOLogger: Instance of a VWOLogger object
    """
    global _logger
    if _logger and kwargs:
        raise Exception("Initializing logger when it is already initialized")
    elif _logger:
        return _logger
    else:
        _logger = VWOLogger(**kwargs)
        return _logger


def create_console_logger(name=__name__, level=None):
    """ Creates a console logger with given name and level using python's standard logging module.

    Args:
        name (str): Name of the logger
        level (logging._levelNames): A logging level accepted by logger

    Returns:
        logging.Logger: A logger instance
    """
    logger = logging.getLogger(name)
    logger.setLevel(level or _DEFAULT_LOGGING_LEVEL)

    handler = logging.StreamHandler()
    handler.setFormatter(logging.Formatter(_VWO_LOG_FORMAT))
    logger.handlers = [handler]

    logger.log(
        logging.DEBUG,
        LogMessageEnum.DEBUG_MESSAGES.LOG_LEVEL_SET.format(file=FILE, level=level).replace("API_NAME", "SDK", 1),
    )
    return logger


class VWOLogger:
    """Module logger class to provide logging functionality. It accepts a logger object with
    log method in it or a log level to create default console logger"""

    def __init__(self, logger=None, log_level=None):
        """ Initializes VWOLogger with provided logger object or creates a console logger object
        with provided level.

        Args:
            logger (CustomLogger|logging.Logger|None): A logger object
            log_level (logging._levelNames): Level of log
        """
        # Set default api name as SDK
        self.api_name = API_NAME
        if logger:
            if isinstance(logger, logging.Logger):
                logger.log(
                    logging.DEBUG,
                    LogMessageEnum.DEBUG_MESSAGES.LOGGING_LOGGER_INSTANCE_USED.format(file=FILE).replace(
                        "API_NAME", self.api_name, 1
                    ),
                )
                self.logger = logger
            else:
                try:
                    logger.log(
                        logging.DEBUG,
                        LogMessageEnum.DEBUG_MESSAGES.CUSTOM_LOGGER_USED.format(file=FILE).replace(
                            "API_NAME", self.api_name, 1
                        ),
                    )
                    self.logger = logger
                except Exception:
                    self.logger = create_console_logger()
                    self.log(
                        logging.ERROR,
                        LogMessageEnum.ERROR_MESSAGES.CUSTOM_LOGGER_MISCONFIGURED.format(
                            file=FILE, extra_info="log method is not provided or invalid."
                        ).replace("API_NAME", self.api_name, 1),
                    )
        else:
            self.logger = create_console_logger(level=log_level)

    def set_api(self, api_name):
        """ It sets the api_name to the API currently being used, for logging purpose. """
        self.api_name = api_name

    def log(self, level, message):
        """ Log method which takes two parameters and logs the message according to
        handler of logger provided while instantiating this class.

        Args:
            level (int): Level of log
            message (string): Message to log
        """

        try:
            self.logger.log(level, message.replace("API_NAME", self.api_name, 1))
        except Exception:
            # Even logging.Logger is broken somehow, simply print to console
            print(level, message)



# Copyright 2019-2021 Wingify Software Pvt. Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

VARIABLE = {
    "type": "object",
    "properties": {
        "id": {"type": ["number", "string"]},
        "key": {"type": ["string"]},
        "value": {"type": ["string", "number", "boolean"]},
        "type": {"type": ["string"]},
    },
    "required": ["id", "value", "key", "type"],
}



# Copyright 2019-2021 Wingify Software Pvt. Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Utility for managing feature related functionalities """

from ..constants import constants
from ..enums.log_message_enum import LogMessageEnum
from ..enums.file_name_enum import FileNameEnum
from ..enums.log_level_enum import LogLevelEnum
from ..logger import VWOLogger

FILE = FileNameEnum.Helpers.FeatureUtil


def get_type_casted_feature_value(value, variable_type):
    """ Returns type casted value to given value type if possbile.

    Args:
        value (int|float|str|bool): Value to type cast
        variable_type (type): Type to which value needs to be casted
    Returns:
        type_casted_value|None : Type casted value if value can be
        typecasted else None
    """

    # Check if type(value) is already equal to required variable_type
    if type(value) == constants.PY_VARIABLE_TYPES.get(variable_type):
        return value
    try:
        if variable_type == constants.VARIABLE_TYPES.STRING:
            return str(value)
        elif variable_type == constants.VARIABLE_TYPES.INTEGER:
            return int(value)
        elif variable_type == constants.VARIABLE_TYPES.DOUBLE:
            return float(value)
        else:
            # required type is boolean, simply raise exception
            # as it doesn't belong to bool type
            raise Exception
    except Exception:
        VWOLogger.getInstance().log(
            LogLevelEnum.ERROR,
            LogMessageEnum.ERROR_MESSAGES.UNABLE_TO_TYPE_CAST.format(
                file=FILE, value=value, variable_type=variable_type, of_type=type(value)
            ),
        )
        return None



# Copyright 2019-2021 Wingify Software Pvt. Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .variable_schema import VARIABLE
from .empty_object_schema import EMPTY_OBJECT


VARIATION = {
    "type": "object",
    "properties": {
        "id": {"type": ["number", "string"]},
        "name": {"type": ["string"]},
        "weight": {"type": ["number", "string"]},
        "variables": {"if": {"type": "array"}, "then": {"items": VARIABLE}, "else": EMPTY_OBJECT},
        "isFeatureEnabled": {"type": "boolean"},
    },
    "required": ["id", "name", "weight"],
}



# Copyright 2019-2021 Wingify Software Pvt. Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from ..helpers import impression_util
from ..constants import constants
from ..logger import VWOLogger
from ..enums.log_level_enum import LogLevelEnum
from ..enums.log_message_enum import LogMessageEnum
from ..enums.file_name_enum import FileNameEnum
import json

FILE = FileNameEnum.Helpers.CustomDimensionsUtil


def get_url_params(settings_file, tag_key, tag_value, user_id):
    """ Creates a property object for custom dimension push api

    Args:
        settings_file (dict): settings_file consisting all the campaign related data
        tag_key (str): str value of length not greater than 255
        tag_value (str): str value of length not greater than 255
        user_id (str): unique user identifier for which custom dimensions should be stored

    Returns:
        dict : dict of all the required properties
    """
    url = constants.HTTPS_PROTOCOL + constants.ENDPOINTS.BASE_URL + constants.ENDPOINTS.PUSH
    tag = {"u": {tag_key: tag_value}}

    params = impression_util.get_common_properties(user_id, settings_file)
    params.update(url=url, tags=json.dumps(tag))

    VWOLogger.getInstance().log(
        LogLevelEnum.DEBUG,
        LogMessageEnum.DEBUG_MESSAGES.PARAMS_FOR_PUSH_CALL.format(file=FILE, properties=json.dumps(params)),
    )
    return params



# Copyright 2019-2021 Wingify Software Pvt. Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..helpers import impression_util
from ..constants import constants
from ..constants.constants import API_METHODS
from ..helpers import campaign_util, validate_util
from ..enums.log_message_enum import LogMessageEnum
from ..enums.file_name_enum import FileNameEnum
from ..enums.log_level_enum import LogLevelEnum

FILE = FileNameEnum.Api.IsFeatureEnabled


def _is_feature_enabled(vwo_instance, campaign_key, user_id, **kwargs):
    """ This API method: Identifies whether the user becomes a part
    of feature rollout/test or not.

    1. Validates the arguments being passed
    2. Checks if user is eligible to get bucketed into the
        feature test/rollout,
    3. Assigns the determinitic variation to the user(based on userId),
        if user becomes part of feature test/rollout
        If UserStorage is used, it will look into it for the
            variation and if found, no further processing is done

    Args:
        campaign_key (string): unique campaign key
        user_id (string): ID assigned to a user

    Keywork Args:
        custom_variables (dict): Custom variables required for segmentation
        variation_targeting_variables (dict): Whitelisting variables to target users

    Returns:
        bool: True if user becomes part of feature test/rollout,
        otherwise false in case user doesn't becomes part of it.
    """

    vwo_instance.logger.set_api(API_METHODS.IS_FEATURE_ENABLED)
    # Retrieve custom variables
    custom_variables = kwargs.get("custom_variables")
    variation_targeting_variables = kwargs.get("variation_targeting_variables")

    if (
        not validate_util.is_valid_string(campaign_key)
        or not validate_util.is_valid_string(user_id)
        or (custom_variables is not None and not validate_util.is_valid_dict(custom_variables))
        or (
            variation_targeting_variables is not None and not validate_util.is_valid_dict(variation_targeting_variables)
        )
    ):  # noqa: E501
        # log invalid params
        vwo_instance.logger.log(
            LogLevelEnum.ERROR, LogMessageEnum.ERROR_MESSAGES.IS_FEATURE_ENABLED_API_INVALID_PARAMS.format(file=FILE)
        )
        return False

    # Get the campaign settings
    campaign = campaign_util.get_campaign(vwo_instance.settings_file, campaign_key)

    # Validate campaign
    if not campaign:
        return False

    # Validate campaign_type
    campaign_type = campaign.get("type")

    if campaign_type == constants.CAMPAIGN_TYPES.VISUAL_AB:
        vwo_instance.logger.log(
            LogLevelEnum.ERROR,
            LogMessageEnum.ERROR_MESSAGES.INVALID_API.format(
                file=FILE, user_id=user_id, campaign_key=campaign_key, campaign_type=campaign_type
            ),
        )
        return False

    # Get variation
    variation = vwo_instance.variation_decider.get_variation(
        user_id,
        campaign,
        custom_variables=custom_variables,
        variation_targeting_variables=variation_targeting_variables,
    )  # noqa: E501

    # If no variation, did not become part of feature_test/rollout
    if not variation:
        return False

    # if campaign type is feature_test Send track call to server
    if campaign_type == constants.CAMPAIGN_TYPES.FEATURE_TEST:
        impression = impression_util.create_impression(
            vwo_instance.settings_file, campaign.get("id"), variation.get("id"), user_id
        )

        vwo_instance.event_dispatcher.dispatch(impression)
        vwo_instance.logger.log(
            LogLevelEnum.INFO,
            LogMessageEnum.INFO_MESSAGES.MAIN_KEYS_FOR_IMPRESSION.format(
                file=FILE,
                campaign_id=impression.get("experiment_id"),
                user_id=impression.get("uId"),
                account_id=impression.get("account_id"),
                variation_id=impression.get("combination"),
            ),
        )
        result = variation.get("isFeatureEnabled")
        if result:
            vwo_instance.logger.log(
                LogLevelEnum.INFO,
                LogMessageEnum.INFO_MESSAGES.FEATURE_ENABLED_FOR_USER.format(
                    file=FILE, user_id=user_id, feature_key=campaign_key
                ),
            )
        else:
            vwo_instance.logger.log(
                LogLevelEnum.INFO,
                LogMessageEnum.INFO_MESSAGES.FEATURE_NOT_ENABLED_FOR_USER.format(
                    file=FILE, user_id=user_id, feature_key=campaign_key
                ),
            )
        return result
    return True



# Copyright 2019-2021 Wingify Software Pvt. Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..helpers import impression_util
from ..constants import constants
from ..constants.constants import API_METHODS
from ..helpers import campaign_util, validate_util
from ..enums.log_message_enum import LogMessageEnum
from ..enums.file_name_enum import FileNameEnum
from ..enums.log_level_enum import LogLevelEnum

FILE = FileNameEnum.Api.Track


def _track(vwo_instance, campaign_specifier, user_id, goal_identifier, **kwargs):
    """
    This API method: Marks the conversion of the campaign(s) for a particular goal

    1. validates the arguments being passed
    2. retrieves the campaigns having the same global goal
    3. calls track_campaign_goal for all the goals

    Args:
        campaign_specifier (None, list, string): Campaign key(s), it can be None in case
        of all campaigns, list in case of given campaigns and string in case of particular
        campaign should to be tracked.
        user_id (string): ID assigned to a user
        goal_identifier (string): campaign(s)'s unique goal identifier

    Keyword Args:
        revenue_value (int|float|string): Provide it through **kwargs.
        It is the revenue generated on triggering the goal
        custom_variables (dict): Custom variables required for segmentation
        variation_targeting_variables (dict): Whitelisting variables to target users

    Returns:
        dict|None: None if called for single campaign and no goal tracked or
        called for all campaigns and no goal tracked.
        Dict otherwise of campaign_key with True/False showing whether the goal
        has been tracked for the campaign or not
    """

    vwo_instance.logger.set_api(API_METHODS.TRACK)
    # Retrive revenue value and custom_variables
    revenue_value = kwargs.get("revenue_value")
    custom_variables = kwargs.get("custom_variables")
    variation_targeting_variables = kwargs.get("variation_targeting_variables")
    valid_params = True
    # Check for valid args
    if (
        not validate_util.is_valid_string(user_id)
        or not validate_util.is_valid_string(goal_identifier)
        or (custom_variables is not None and not validate_util.is_valid_dict(custom_variables))
        or (
            variation_targeting_variables is not None and not validate_util.is_valid_dict(variation_targeting_variables)
        )
        or (revenue_value is not None and not validate_util.is_valid_basic_data_type(revenue_value))
    ):
        valid_params = False

    goal_type_to_track = kwargs.get("goal_type_to_track")
    if goal_type_to_track is None:
        goal_type_to_track = vwo_instance.goal_type_to_track
    elif not validate_util.is_valid_goal_type(goal_type_to_track):
        valid_params = False
    should_track_returning_user = kwargs.get("should_track_returning_user")
    if should_track_returning_user is None:
        should_track_returning_user = vwo_instance.should_track_returning_user
    elif type(should_track_returning_user) is not bool:
        valid_params = False

    if not valid_params:
        vwo_instance.logger.log(
            LogLevelEnum.ERROR, LogMessageEnum.ERROR_MESSAGES.TRACK_API_INVALID_PARAMS.format(file=FILE)
        )
        return None

    campaigns_without_goal = []
    no_campaign_found = False
    if type(campaign_specifier) is str:
        campaign = campaign_util.get_campaign(vwo_instance.settings_file, campaign_specifier)
        goal = campaign_util.get_campaign_goal(campaign, goal_identifier)
        if not goal:
            no_campaign_found = True
        else:
            campaign_goal_list = [(campaign, goal)]
    elif type(campaign_specifier) is list:
        campaigns = campaign_util.get_campaigns(vwo_instance.settings_file, campaign_specifier).values()
        (campaign_goal_list, campaigns_without_goal) = campaign_util.get_campaigns_with_goal_id(
            campaigns, goal_identifier
        )
        for campaign in campaigns_without_goal:
            vwo_instance.logger.log(
                LogLevelEnum.ERROR,
                LogMessageEnum.ERROR_MESSAGES.TRACK_API_GOAL_NOT_FOUND.format(
                    file=FILE, goal_identifier=goal_identifier, user_id=user_id, campaign_key=campaign.get("key")
                ),
            )
    elif campaign_specifier is None:
        campaigns = vwo_instance.settings_file.get("campaigns")
        campaign_goal_list = campaign_util.get_campaigns_with_goal_id(campaigns, goal_identifier)[0]
        if not campaign_goal_list:
            no_campaign_found = True
    else:
        vwo_instance.logger.log(
            # Specific log for campaign_specifier type
            LogLevelEnum.ERROR,
            LogMessageEnum.ERROR_MESSAGES.TRACK_API_INVALID_PARAMS.format(file=FILE),
        )
        return None

    if no_campaign_found:
        vwo_instance.logger.log(
            LogLevelEnum.ERROR,
            LogMessageEnum.ERROR_MESSAGES.NO_CAMPAIGN_FOUND.format(file=FILE, goal_identifier=goal_identifier),
        )
        return None

    ret_value = {}
    for campaign, goal in campaign_goal_list:
        result = track_campaign_goal(
            vwo_instance,
            campaign,
            user_id,
            goal,
            revenue_value,
            custom_variables,
            variation_targeting_variables,
            goal_type_to_track,
            should_track_returning_user,
        )
        ret_value[campaign.get("key")] = result
    for campaign in campaigns_without_goal:
        ret_value[campaign.get("key")] = False
    return ret_value


def track_campaign_goal(
    vwo_instance,
    campaign,
    user_id,
    goal,
    revenue_value,
    custom_variables,
    variation_targeting_variables,
    goal_type_to_track,
    should_track_returning_user,
):
    """
    It marks the conversion of given goal for the given campaign

    1. Checks if user is eligible to get bucketed into the campaign,
    2. Gets the assigned determinitic variation to the
        user(based on userId), if user becomes part of campaign
    3. Sends an impression call to VWO server to track goal data

    Args:
        campaign (dict): Campaign object
        user_id (string): ID assigned to a user
        goal (dict): Goal object
        revenue_value (int|float|string): It is the revenue generated on triggering the goal
        custom_variables (dict): Custom variables required for segmentation
        variation_targeting_variables (dict): Whitelisting variables to target users
        goal_type_to_track (vwo.GOAL_TYPES): Goal type that should be tracked in case of mixed
        global goal identifier
        should_track_returning_user (bool): Should a returning user be tracked or not, user_storage
        is required for this to work

    Returns:
        bool: True if goal successfully tracked else False
    """

    campaign_type = campaign.get("type")
    if campaign_type == constants.CAMPAIGN_TYPES.FEATURE_ROLLOUT:
        vwo_instance.logger.log(
            LogLevelEnum.ERROR,
            LogMessageEnum.ERROR_MESSAGES.INVALID_API.format(
                file=FILE, user_id=user_id, campaign_key=campaign.get("key"), campaign_type=campaign_type
            ),
        )
        return False

    goal_type = goal.get("type")

    if (goal_type_to_track == constants.GOAL_TYPES.CUSTOM and goal_type == constants.GOAL_TYPES.REVENUE) or (
        goal_type_to_track == constants.GOAL_TYPES.REVENUE and goal_type == constants.GOAL_TYPES.CUSTOM
    ):
        # We can log goal type didn't match in debug mode
        return False

    if goal_type == constants.GOAL_TYPES.REVENUE and not validate_util.is_valid_value(revenue_value):
        vwo_instance.logger.log(
            LogLevelEnum.ERROR,
            LogMessageEnum.ERROR_MESSAGES.TRACK_API_REVENUE_NOT_PASSED_FOR_REVENUE_GOAL.format(
                file=FILE, user_id=user_id, goal_identifier=goal.get("identifier"), campaign_key=campaign.get("key")
            ),
        )
        return False

    if goal_type == constants.GOAL_TYPES.CUSTOM:
        revenue_value = None

    variation = vwo_instance.variation_decider.get_variation(
        user_id,
        campaign,
        custom_variables=custom_variables,
        variation_targeting_variables=variation_targeting_variables,
        goal_data={"identifier": goal.get("identifier"), "should_track_returning_user": should_track_returning_user},
    )

    if variation:
        impression = impression_util.create_impression(
            vwo_instance.settings_file, campaign.get("id"), variation.get("id"), user_id, goal.get("id"), revenue_value
        )

        vwo_instance.event_dispatcher.dispatch(impression)

        vwo_instance.logger.log(
            LogLevelEnum.INFO,
            LogMessageEnum.INFO_MESSAGES.MAIN_KEYS_FOR_IMPRESSION.format(
                file=FILE,
                campaign_id=impression.get("experiment_id"),
                user_id=impression.get("uId"),
                account_id=impression.get("account_id"),
                variation_id=impression.get("combination"),
            ),
        )
        return True
    return False



# Copyright 2019-2021 Wingify Software Pvt. Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..constants import constants
from ..constants.constants import API_METHODS
from ..helpers import campaign_util, validate_util, feature_util
from ..enums.log_message_enum import LogMessageEnum
from ..enums.file_name_enum import FileNameEnum
from ..enums.log_level_enum import LogLevelEnum

FILE = FileNameEnum.Api.GetFeatureVariableValue


def _get_feature_variable_value(vwo_instance, campaign_key, variable_key, user_id, **kwargs):
    """ Returns the feature variable corresponding to the variable_key
    passed. It typecasts the value to the corresponding value type
    found in settings_file

    1. Validates the arguments being passed
    2. Checks if user is eligible to get bucketed into the feature test/rollout,
    3. Assigns the determinitic variation to the user(based on userId),
        if user becomes part of campaign
        If UserStorage is used, it will look into it for the
            variation and if found, no further processing is done
    4. Retrieves the corresponding variable from variation assigned.

    Args:
        campaign_key (string): unique campaign key
        variable_key (string): variable key
        user_id (string): ID assigned to a user

    Keywork Args:
        custom_variables (dict): Custom variables required for segmentation
        variation_targeting_variables (dict): Whitelisting variables to target users

    Returns:
        variable(bool, str, int, float)|None: If variation is assigned then
        variable corresponding to variation assigned else None
    """

    vwo_instance.logger.set_api(API_METHODS.GET_FEATURE_VARIABLE_VALUE)
    # Retrieve custom variables
    custom_variables = kwargs.get("custom_variables")
    variation_targeting_variables = kwargs.get("variation_targeting_variables")

    if (
        not validate_util.is_valid_string(campaign_key)
        or not validate_util.is_valid_string(variable_key)
        or not validate_util.is_valid_string(user_id)
        or (custom_variables is not None and not validate_util.is_valid_dict(custom_variables))
        or (
            variation_targeting_variables is not None and not validate_util.is_valid_dict(variation_targeting_variables)
        )
    ):  # noqa: E501
        vwo_instance.logger.log(
            LogLevelEnum.ERROR,
            LogMessageEnum.ERROR_MESSAGES.GET_FEATURE_VARIABLE_VALUE_API_INVALID_PARAMS.format(file=FILE),
        )
        return None

    # Get the campaign settings
    campaign = campaign_util.get_campaign(vwo_instance.settings_file, campaign_key)

    # Validate campaign
    if not campaign:
        return None

    campaign_type = campaign.get("type")

    if campaign_type == constants.CAMPAIGN_TYPES.VISUAL_AB:
        vwo_instance.logger.log(
            LogLevelEnum.ERROR,
            LogMessageEnum.ERROR_MESSAGES.INVALID_API.format(
                file=FILE, campaign_key=campaign_key, campaign_type=campaign_type, user_id=user_id
            ),
        )
        return None

    variation = vwo_instance.variation_decider.get_variation(
        user_id,
        campaign,
        custom_variables=custom_variables,
        variation_targeting_variables=variation_targeting_variables,
    )  # noqa: E501

    # Check if variation has been assigned to user
    if not variation:
        return None

    # Variation received, evaluate variable
    variable = None

    if campaign_type == constants.CAMPAIGN_TYPES.FEATURE_ROLLOUT:
        variables = campaign.get("variables")

    elif campaign_type == constants.CAMPAIGN_TYPES.FEATURE_TEST:
        if variation.get("isFeatureEnabled") is False:
            vwo_instance.logger.log(
                LogLevelEnum.INFO,
                LogMessageEnum.INFO_MESSAGES.FEATURE_NOT_ENABLED_FOR_USER.format(
                    file=FILE, feature_key=campaign_key, user_id=user_id
                ),
            )
            variation = campaign_util.get_control_variation(campaign)
        else:
            vwo_instance.logger.log(
                LogLevelEnum.INFO,
                LogMessageEnum.INFO_MESSAGES.FEATURE_ENABLED_FOR_USER.format(
                    file=FILE, feature_key=campaign_key, user_id=user_id
                ),
            )
        variables = variation.get("variables")

    variable = campaign_util.get_variable(variables, variable_key)

    if not variable:
        # Log variable not found
        vwo_instance.logger.log(
            LogLevelEnum.ERROR,
            LogMessageEnum.ERROR_MESSAGES.VARIABLE_NOT_FOUND.format(
                file=FILE,
                variable_key=variable_key,
                campaign_key=campaign_key,
                campaign_type=campaign_type,
                user_id=user_id,
            ),
        )
        return None

    vwo_instance.logger.log(
        LogLevelEnum.INFO,
        LogMessageEnum.INFO_MESSAGES.VARIABLE_FOUND.format(
            file=FILE,
            variable_key=variable_key,
            variable_value=variable.get("value"),
            campaign_key=campaign_key,
            campaign_type=campaign_type,
            user_id=user_id,
        ),
    )

    return feature_util.get_type_casted_feature_value(variable.get("value"), variable.get("type"))



# Copyright 2019-2021 Wingify Software Pvt. Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import re

from ...helpers.generic_util import get_key_value
from ...helpers.segment_utils import convert_to_true_types, process_custom_variables_value, process_operand_value


class OperandEvaluator:
    """ Class to evaluate different kinds of operands supported by segmentation """

    def lower(self, operand_value, custom_variables_value):
        """ Checks if both values are same after 'lowercasing'
            i.e. case insensitive check

        Args:
            operand_value: Leaf value from the segments
            custom_variables_value: Value from the custom_variables

        Returns:
            bool (result): True or False
        """
        return operand_value.lower() == custom_variables_value.lower()

    def contains(self, operand_value, custom_variables_value):
        """ Checks if custom_variables_value contains operand_value

        Args:
            operand_value: Leaf value from the segments
            custom_variables_value: Value from the custom_variables

        Returns:
            bool (result): True or False
        """
        return custom_variables_value.__contains__(operand_value)

    def startswith(self, operand_value, custom_variables_value):
        """ Checks if custom_variables_value starts with operand_value

        Args:
            operand_value: Leaf value from the segments
            custom_variables_value: Value from the custom_variables

        Returns:
            bool (result): True or False
        """
        return custom_variables_value.endswith(operand_value)

    def endswith(self, operand_value, custom_variables_value):
        """ Checks if custom_variables_value ends with operand_value

        Args:
            operand_value: Leaf value from the segments
            custom_variables_value: Value from the custom_variables

        Returns:
            bool (result): True or False
        """
        return custom_variables_value.startswith(operand_value)

    def regex(self, operand_value, custom_variables_value):
        """ Checks if custom_variables_value matches the regex specified by
            operand_value

        Args:
            operand_value: Leaf value from the segments
            custom_variables_value: Value from the custom_variables

        Returns:
            bool (result): True or False
        """
        try:
            pattern = re.compile(operand_value)
            return bool(pattern.search(custom_variables_value))
        except Exception:
            return False

    def equals(self, operand_value, custom_variables_value):
        """ Checks if both values are exactly same

        Args:
            operand_value: Leaf value from the segments
            custom_variables_value: Value from the custom_variables

        Returns:
            bool (result): True or False
        """
        return custom_variables_value == operand_value

    def evaluate_custom_variable(self, operand, custom_variables):
        """ Identifies the condition stated in the leaf node and evaluates the result

        Args:
            operand(str): String representation of operand_type and operand_value,
            for eg. lower(vwo), wildcard(www.vwo.com/*), etc.
            custom_variables(dict): Custom variables provided to the sdk through the api

        Returns:
            boolean (result): True if custom_variable satisfies condition stated by operand_value,
            else False
        """

        # Extract custom_variable_key and custom_variables_value from operand
        operand_key, operand = get_key_value(operand)

        # Retrieve corresponding custom_variable value from custom_variables
        custom_variables_value = custom_variables.get(operand_key)

        # Pre process custom_variable value
        custom_variables_value = process_custom_variables_value(custom_variables_value)

        # Pre process operand value
        operand_type, operand_value = process_operand_value(operand)

        # Process the custom_variables_value and operand_value to make them of same type
        operand_value, custom_variables_value = convert_to_true_types(operand_value, custom_variables_value)

        # Call the self method corresponding to operand_type to evaluate the result
        return getattr(self, operand_type)(operand_value, custom_variables_value)

    def evaluate_user(self, operand, variation_targeting_variables):
        """ Identifies the condition stated in the leaf node and evaluates the result

        Args:
            operand(str): String representation of comma separated user_ids
            variation_targeting_variables(dict): Variation targeting variables provided to the sdk through the api

        Returns:
            boolean (result): True if _vwo_user_id is in comma separated user_ids represented in operand
            else False
        """

        # Retrieve corresponding _vwo_user_id value from variation_targeting_variables
        _vwo_user_id = variation_targeting_variables.get("_vwo_user_id")

        # Extract user_id list from operand
        operand_user_id_list = operand.split(",")

        # check if _vwo_user_id exists in operand_user_id_list or not
        for user_id in operand_user_id_list:
            if user_id.strip() == _vwo_user_id:
                return True
        return False



# Copyright 2019-2021 Wingify Software Pvt. Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division
import mmh3 as Hasher
from ..constants import constants
from ..helpers import validate_util
from ..enums.log_message_enum import LogMessageEnum
from ..enums.file_name_enum import FileNameEnum
from ..enums.log_level_enum import LogLevelEnum
from ..logger import VWOLogger

# Took reference from StackOverflow(https://stackoverflow.com/) to:
# convert signed to unsigned integer in python from StackOverflow
# Author - Duncan (https://stackoverflow.com/users/107660/duncan)
# Source - https://stackoverflow.com/a/20766900/2494535
U_MAX_32_BIT = 0xFFFFFFFF
FILE = FileNameEnum.Core.Bucketer


class Bucketer(object):
    """ Class consisting the core bucketing/distribution logic for
    the visitors """

    def __init__(self):
        """ Initializes bucketer with vwo common logger """
        self.logger = VWOLogger.getInstance()

    def get_variation(self, variations, bucket_value):
        """ Returns the Variation by checking the Start and End
        Bucket Allocations of each Variation

        Args:
            variations (list): list of variations
            bucket_value (int): the bucket value of the user

        Returns:
            (dict|None): variation data allotted to the user or None if not
        """
        for variation in variations:
            if variation.get("start_variation_allocation") <= bucket_value <= variation.get("end_variation_allocation"):
                return variation
        return None

    def get_bucket_value_for_user(self, user_id, max_value, multiplier=1):
        """ Returns Bucket Value of the user by hashing the userId with murmur hash and scaling it down.

        Args:
            user_id (string): the unique ID assigned to User
            max_value(int): maximum value that can be alloted to the bucket value
            multiplier(int): value for distributing ranges slightly
        Returns:
            int: the bucket value allotted to User
            (between 1 to MAX_TRAFFIC_PERCENT)
        """

        hash_value = Hasher.hash(user_id, constants.SEED_VALUE) & U_MAX_32_BIT
        ratio = hash_value / (2 ** 32)
        multiplied_value = (max_value * ratio + 1) * multiplier
        bucket_value = int(multiplied_value)

        self.logger.log(
            LogLevelEnum.DEBUG,
            LogMessageEnum.DEBUG_MESSAGES.USER_HASH_BUCKET_VALUE.format(
                file=FILE, hash_value=hash_value, bucket_value=bucket_value, user_id=user_id
            ),
        )
        return bucket_value

    def is_user_part_of_campaign(self, user_id, campaign):
        """ Calculates if the provided user_id should become part of the campaign or not

        Args:
            user_id (strings): the unique ID assigned to a user
            campaign (dict): for getting traffic allotted to the campaign

        Returns:
            bool: if User is a part of Campaign or not
        """

        if not validate_util.is_valid_value(user_id):
            self.logger.log(
                LogLevelEnum.ERROR,
                LogMessageEnum.ERROR_MESSAGES.INVALID_USER_ID.format(
                    file=FILE, user_id=user_id, method="is_user_part_of_campaign"
                ),
            )
            return False

        if not campaign:
            self.logger.log(
                LogLevelEnum.ERROR,
                LogMessageEnum.ERROR_MESSAGES.INVALID_CAMPAIGN.format(file=FILE, method="is_user_part_of_campaign"),
            )
            return False

        traffic_allocation = campaign.get("percentTraffic")

        value_assigned_to_user = self.get_bucket_value_for_user(user_id, constants.MAX_TRAFFIC_PERCENT)
        is_user_part = value_assigned_to_user != 0 and value_assigned_to_user <= traffic_allocation
        self.logger.log(
            LogLevelEnum.INFO,
            LogMessageEnum.INFO_MESSAGES.USER_ELIGIBILITY_FOR_CAMPAIGN.format(
                file=FILE, user_id=user_id, is_user_part=is_user_part
            ),
        )
        return is_user_part

    def bucket_user_to_variation(self, user_id, campaign):
        """ Validates the User ID and
            returns Variation into which the User is bucketed in.

        Args:
            user_id (string): the unique ID assigned to User
            campaign (dict): the Campaign of which User is a part of

        Returns:
            (dict|None): variation data into which user is bucketed in
                or None if not
        """

        if not validate_util.is_valid_value(user_id):
            self.logger.log(
                LogLevelEnum.ERROR,
                LogMessageEnum.ERROR_MESSAGES.INVALID_USER_ID.format(
                    file=FILE, user_id=user_id, method="bucket_user_to_variation"
                ),
            )
            return None

        if not campaign:
            self.logger.log(
                LogLevelEnum.ERROR,
                LogMessageEnum.ERROR_MESSAGES.INVALID_CAMPAIGN.format(file=FILE, method="bucket_user_to_variation"),
            )
            return None

        normalize = constants.MAX_TRAFFIC_VALUE / campaign.get("percentTraffic")
        multiplier = normalize / 100
        bucket_value = self.get_bucket_value_for_user(user_id, constants.MAX_TRAFFIC_VALUE, multiplier)

        self.logger.log(
            LogLevelEnum.DEBUG,
            LogMessageEnum.DEBUG_MESSAGES.VARIATION_HASH_BUCKET_VALUE.format(
                file=FILE,
                user_id=user_id,
                campaign_key=campaign.get("key"),
                percent_traffic=campaign.get("percentTraffic"),
                bucket_value=bucket_value,
            ),
        )
        return self.get_variation(campaign.get("variations"), bucket_value)



# Copyright 2019-2021 Wingify Software Pvt. Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

GOAL = {
    "items": {
        "type": "object",
        "properties": {"identifier": {"type": "string"}, "id": {"type": "number"}, "type": {"type": "string"}},
        "required": ["identifier", "id", "type"],
    }
}



# Copyright 2019-2021 Wingify Software Pvt. Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Module for making requests, uses requests internally """

import requests


class Connection:
    """ Connection class to provide SDK with network connectivity interfaces """

    def __init__(self):
        """ Initializes connection class with requests session object"""
        self.session = requests.Session()

    def get(self, url, params=None):
        """ Get method, it wraps upon requests' get method.
        Args:
            url (str): Unique resource locator
            params (dict): Parameters to be passed
        Returns:
            dict : Status code and Response text
        """
        try:
            resp = self.session.get(url, params=params)
            return {"status_code": resp.status_code, "text": resp.text}
        except Exception:
            return {"status_code": None, "text": ""}

    def post(self, url, params=None, data=None):
        """ Post method, it wraps upon requests' post method.
        Args:
            url (str): Unique resource locator
            params (dict): Parameters to be passed
            data (dict): Json data to be passed
        Returns:
            dict : Status code and Response text
        """
        try:
            resp = self.session.post(url, params=params, json=data)
            return {"status_code": resp.status_code, "text": resp.text}
        except Exception:
            return {"status_code": None, "text": ""}



# Copyright 2019-2021 Wingify Software Pvt. Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# flake8: noqa

""" Schema for verifying the settings_file provided by the customer """

from .campaign_schema import CAMPAIGN
from .empty_object_schema import EMPTY_OBJECT


SETTINGS_FILE_SCHEMA = {
    "type": "object",
    "properties": {
        "version": {"type": ["number", "string"]},
        "accountId": {"type": ["number", "string"]},
        "campaigns": {"if": {"type": "array"}, "then": {"items": CAMPAIGN}, "else": EMPTY_OBJECT},
    },
    "required": ["version", "accountId", "campaigns"],
}



# Copyright 2019-2021 Wingify Software Pvt. Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import api
from .core.variation_decider import VariationDecider
from .enums.file_name_enum import FileNameEnum
from .enums.log_level_enum import LogLevelEnum
from .enums.log_message_enum import LogMessageEnum
from .event.event_dispatcher import EventDispatcher
from .helpers.generic_util import safe_method
from .logger import VWOLogger
from .services.settings_file_manager import SettingsFileManager
from .constants.constants import GOAL_TYPES


FILE = FileNameEnum.Vwo.VWO
logger = VWOLogger.getInstance()


class VWO(object):
    """ Core class of the SDK, consisting all the APIs featured in VWO Full Stack Server Side Testing """

    def __init__(
        self, settings_file, user_storage, is_development_mode, goal_type_to_track, should_track_returning_user
    ):
        """ __init__ method to initialize the VWO object, all the argument types should be pre-checked.
        Else object initialization fails.

        Args:
            settings_file (dict): settings_file consisting all the campaign related data
            user_storage (object): a storage service object capable of doing get and set on
            SDK provide data
            is_development_mode (bool): should the SDK be initialized in development mode,
            it toggles the event_dispatcher to off
            goal_type_to_track (vwo.GOAL_TYPES): which goal type to track when using track
            api. Default value is vwo.GOAL_TYPES.ALL
            should_track_returning_user (bool): should returning user be tracked again.
            Default value is False
        """
        self.logger = VWOLogger.getInstance()
        self.config = SettingsFileManager(settings_file)
        self.settings_file = self.config.get_settings_file()
        self.variation_decider = VariationDecider(user_storage)
        if is_development_mode:
            self.logger.log(LogLevelEnum.DEBUG, LogMessageEnum.DEBUG_MESSAGES.SET_DEVELOPMENT_MODE.format(file=FILE))
        self.event_dispatcher = EventDispatcher(is_development_mode or False)
        self.goal_type_to_track = goal_type_to_track or GOAL_TYPES.ALL
        self.should_track_returning_user = should_track_returning_user or False
        self.logger.log(LogLevelEnum.DEBUG, LogMessageEnum.DEBUG_MESSAGES.SDK_INITIALIZED.format(file=FILE))

    # PUBLIC METHODS
    activate = safe_method(api._activate, None, FILE)
    get_variation_name = safe_method(api._get_variation_name, None, FILE)
    track = safe_method(api._track, False, FILE)
    is_feature_enabled = safe_method(api._is_feature_enabled, False, FILE)
    get_feature_variable_value = safe_method(api._get_feature_variable_value, None, FILE)
    push = safe_method(api._push, False, FILE)



# Copyright 2019-2021 Wingify Software Pvt. Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class LogMessageEnum:
    """ Classobj encapsulating various logging messages """

    class DEBUG_MESSAGES:
        """ Classobj encapsulating various DEBUG messages """

        LOG_LEVEL_SET = "({file}): [API_NAME] Log level set to {level}"
        SET_DEVELOPMENT_MODE = "({file}): [API_NAME] DEVELOPMENT mode is ON"
        VALID_CONFIGURATION = "({file}): [API_NAME] SDK configuration and account settings are valid."
        CUSTOM_LOGGER_USED = "({file}): [API_NAME] Custom logger used"
        LOGGING_LOGGER_INSTANCE_USED = "({file}): [API_NAME] Python logging module's logger instantiated"
        SDK_INITIALIZED = "({file}): [API_NAME] SDK properly initialzed"
        SETTINGS_FILE_PROCESSED = "({file}): [API_NAME] Settings file processed"
        NO_STORED_VARIATION = "({file}): [API_NAME] No stored variation for user_id:{user_id} for campaign_key:{campaign_key} found in UserStorage"
        NO_USER_STORAGE_GET = "({file}): [API_NAME] No UserStorage to get data"
        NO_USER_STORAGE_SET = "({file}): [API_NAME] No UserStorage to set data"
        GETTING_STORED_VARIATION = "({file}): [API_NAME] Got stored variation for user_id:{user_id} of campaign_key:{campaign_key} as variation: {variation_name} found in UserStorage"
        USER_HASH_BUCKET_VALUE = (
            "({file}): [API_NAME] user_id:{user_id} having hash:{hash_value} got bucketValue:{bucket_value}"
        )
        VARIATION_HASH_BUCKET_VALUE = "({file}): [API_NAME] user_id:{user_id} for campaign_key:{campaign_key} having percent traffic:{percent_traffic} got bucket value:{bucket_value}"
        GOT_VARIATION_FOR_USER = "({file}): [API_NAME] user_id:{user_id} for campaign_key:{campaign_key} type: {campaign_type} got variation_name:{variation_name} inside method:{method}"
        USER_NOT_PART_OF_CAMPAIGN = "({file}): [API_NAME] user_id:{user_id} for campaign_key:{campaign_key} type: {campaign_type} did not become part of campaign method:{method}"
        UUID_FOR_USER = (
            "({file}): [API_NAME] Uuid generated for user_id:{user_id} and account_id:{account_id} is {desired_uuid}"
        )
        IMPRESSION_FOR_TRACK_USER = "({file}): [API_NAME] Impression built for track-user - {properties}"
        IMPRESSION_FOR_TRACK_GOAL = "({file}): [API_NAME] Impression built for track-goal - {properties}"

        PARAMS_FOR_PUSH_CALL = "({file}): [API_NAME] Params for push call - {properties}"

        NO_VARIABLES = "({file}): [API_NAME] For user_id:{user_id} of campaign: {campaign_key} variables are not passed for {segmentation_type}"
        SEGMENTATION_SKIPPED = "({file}): [API_NAME] For user_id:{user_id} of campaign_key: {campaign_key} segments are missing, hence skipping segmentation {variation_status}"
        SEGMENTATION_STATUS = "({file}): [API_NAME] For user_id: {user_id} of campaign_key:{campaign_key} with variables:{variables} {status} {segmentation_type} {variation_status}"

        WHITELISTING_SKIPPED = (
            "({file}): [API_NAME] For user_id:{user_id} of campaign:{campaign_key}, whitelisting is skipped"
        )

    class INFO_MESSAGES:
        """ Classobj encapsulating various INFO messages """

        VARIATION_RANGE_ALLOCATION = "({file}): [API_NAME] campaign_key:{campaign_key} having variations:{variation_name} with weight:{variation_weight} got range as: ( {start} - {end} ))"
        VARIATION_ALLOCATED = "({file}): [API_NAME] user_id:{user_id} of campaign_key:{campaign_key} type: {campaign_type} got variation: {variation_name}"
        FORCED_VARIATION_ALLOCATED = "({file}): [API_NAME] user_id:{user_id} of campaign_key:{campaign_key} type: {campaign_type} got forced-variation: {variation_name}"
        LOOKING_UP_USER_STORAGE = "({file}): [API_NAME] Looked into UserStorage for user_id:{user_id} successful"
        SAVING_DATA_USER_STORAGE_STATUS = "({file}): [API_NAME] Saving into UserStorage for user_id:{user_id} {status}"
        GOT_STORED_VARIATION = "({file}): [API_NAME] Got stored variation:{variation_name} of campaign_key:{campaign_key} for user_id:{user_id} from UserStorage"
        NO_VARIATION_ALLOCATED = "({file}): [API_NAME] user_id:{user_id} of campaign_key:{campaign_key} type: {campaign_type} did not get any variation"
        USER_ELIGIBILITY_FOR_CAMPAIGN = "({file}): [API_NAME] Is user_id:{user_id} part of campaign? {is_user_part}"
        GOT_VARIATION_FOR_USER = "({file}): [API_NAME] user_id:{user_id} for campaign_key:{campaign_key} type: {campaign_type} got variation_name:{variation_name}"
        USER_GOT_NO_VARIATION = "({file}): [API_NAME] user_id:{user_id} for campaign_key:{campaign_key} type: {campaign_type} did not allot any variation"
        IMPRESSION_SUCCESS = "({file}): [API_NAME] Impression event - {end_point} was successfully received by VWO"
        MAIN_KEYS_FOR_IMPRESSION = "({file}): [API_NAME] Having main keys: account_id:{account_id} user_id:{user_id} campaignId:{campaign_id} and variation_id:{variation_id}"
        MAIN_KEYS_FOR_PUSH_API = (
            "({file}): [API_NAME] Having main keys: account_id:{account_id} user_id:{user_id} u:{u} and tags:{tags}"
        )
        INVALID_VARIATION_KEY = (
            "({file}): [API_NAME] variation was not assigned to user_id:{user_id} for campaign_key:{campaign_key}"
        )
        RETRY_FAILED_IMPRESSION_AFTER_DELAY = "({file}): [API_NAME] Failed impression event for {end_point} will be retried after {retry_timeout} milliseconds delay"

        USER_IN_FEATURE_ROLLOUT = "({file}): [API_NAME] User ID:{user_id} is in feature rollout:{campaign_key}"
        USER_NOT_IN_FEATURE_ROLLOUT = "({file}): [API_NAME] User ID:{user_id} is NOT in feature rollout:{campaign_key}"
        FEATURE_ENABLED_FOR_USER = (
            "({file}): [API_NAME] Feature having feature-key:{feature_key} for user ID:{user_id} is enabled"
        )
        FEATURE_NOT_ENABLED_FOR_USER = (
            "({file}): [API_NAME] Feature having feature-key:{feature_key} for user ID:{user_id} is not enabled"
        )

        VARIABLE_FOUND = "({file}): [API_NAME] Value for variable:{variable_key} of campaign_key:{campaign_key} and campaign type: {campaign_type} is:{variable_value} for user:{user_id}"

        SEGMENTATION_STATUS = "({file}): [API_NAME] user_id:{user_id} of campaign_key:{campaign_key} with variables: {variables} {status} {segmentation_type} {variation_status}"

        GOAL_ALREADY_TRACKED = "({file}): [API_NAME] Goal:{goal_identifier} of Campaign:{campaign_key} for User ID:{user_id} has already been tracked earlier. Skipping now. Check docs on how to force tracking for returning visitors."

    class WARNING_MESSAGES:
        """ Classobj encapsulating various WARNING messages """

    class ERROR_MESSAGES:
        """ Classobj encapsulating various ERROR messages """

        LAUNCH_API_INVALID_PARAMS = (
            "({file}): [API_NAME] Params passed to the launch API are not correct. Please refer the docs.."
        )
        ACTIVATE_API_INVALID_PARAMS = "({file}): [API_NAME] API got bad parameters. It expects campaign_key(String) as first and user_id(String) as second argument, custom_variables(dict) for pre-segmentation and variation_targeting_variables(dict) for white-listing can be passed via kwargs"
        API_CONFIG_CORRUPTED = "({file}): [API_NAME] API has corrupted configuration"
        GET_VARIATION_NAME_API_INVALID_PARAMS = "({file}): [API_NAME] API got bad parameters. It expects campaign_key(String) as first and user_id(String) as second argument, custom_variables(dict) for pre-segmentation and variation_targeting_variables(dict) for white-listing can be passed via kwargs"
        TRACK_API_INVALID_PARAMS = "({file}): [API_NAME] API got bad parameters. It expects campaign_key(String or Array of Strings or None) as first user_id(String) as second and goal_identifier(String/Number) as third argument. revenue_value(Float/Number/String) can be passed through kwargs and is required for revenue goal only. custom_variables(dict) for pre-segmentation and variation_targeting_variables(dict) for white-listing can be passed via kwargs"
        TRACK_API_GOAL_NOT_FOUND = "({file}): [API_NAME] Goal:{goal_identifier} not found for campaign_key:{campaign_key} and user_id:{user_id}"
        TRACK_API_REVENUE_NOT_PASSED_FOR_REVENUE_GOAL = "({file}): [API_NAME] Revenue value should be passed for revenue goal:{goal_identifier} for campaign_key:{campaign_key} and user_id:{user_id}"
        TRACK_API_VARIATION_NOT_FOUND = (
            "({file}): [API_NAME] variation not found for campaign_key:{campaign_key} and user_id:{user_id}"
        )
        CAMPAIGN_NOT_RUNNING = (
            "({file}): [API_NAME] campaign_key:{campaign_key} is not RUNNING. Please verify from VWO App"
        )
        LOOK_UP_USER_STORAGE_FAILED = "({file}): [API_NAME] Looking data from UserStorage failed for user_id:{user_id}"
        SET_USER_STORAGE_FAILED = "({file}): [API_NAME] Error while saving data into UserStorage for user_id:{user_id}. Error message: {error_message}"
        INVALID_CAMPAIGN = "({file}): [API_NAME] Invalid campaign passed to {method} of this file"
        INVALID_USER_ID = "({file}): [API_NAME] Invalid user_id:{user_id} passed to {method} of this file"
        IMPRESSION_FAILED = "({file}): [API_NAME] Impression event could not be sent to VWO - {end_point}"
        CUSTOM_LOGGER_MISCONFIGURED = "({file}): [API_NAME] Custom logger is provided but seems to have misconfigured. {extra_info} Please check the API Docs. Using default logger."

        INVALID_API = "({file}): [API_NAME] API is not valid for user ID: {user_id} in campaign ID: {campaign_key} having campaign type: {campaign_type}."
        IS_FEATURE_ENABLED_API_INVALID_PARAMS = "({file}): [API_NAME] API got bad parameters. It expects campaign_key(String) as first and user_id(String) as second argument, custom_variables(dict) for pre-segmentation and variation_targeting_variables(dict) for white-listing can be passed via kwargs"
        GET_FEATURE_VARIABLE_VALUE_API_INVALID_PARAMS = "({file}): [API_NAME] API got bad parameters. It expects campaign_key(String) as first, variable_key(string) as second and user_id(String) as third argument, custom_variables(dict) for pre-segmentation and variation_targeting_variables(dict) for white-listing can be passed via kwargs"

        VARIABLE_NOT_FOUND = "({file}): [API_NAME] Variable {variable_key} not found for campaign {campaign_key} and type {campaign_type} for user ID {user_id}"
        UNABLE_TO_TYPE_CAST = (
            "({file}): [API_NAME] Unable to typecast value: {value} of type: {of_type} to type: {variable_type}."
        )

        USER_NOT_IN_CAMPAIGN = "({file}): [API_NAME] user_id:{user_id} did not become part of campaign_key:{campaign_key} and campaign type:{campaign_type}"
        API_NOT_WORKING = (
            "({file}): [API_NAME] API not working, exception caught: {exception}. Please contact VWO Support for help."
        )

        SEGMENTATION_ERROR = "({file}): [API_NAME] Error while segmenting the user_id:{user_id} of campaign_key:{campaign_key} with variables:{variables}{variation_status}. Error message: {error_message}"

        PUSH_API_INVALID_PARAMS = "({file}): [API_NAME] API got bad parameters. It expects tag_key(String) as first and tag_value(String) as second argument and user_id(String) as third argument"
        TAG_VALUE_LENGTH_EXCEEDED = "({file}): [API_NAME] The length of tag_value:{tag_value} and user_id: {user_id} can not be greater than 255"
        TAG_KEY_LENGTH_EXCEEDED = (
            "({file}): [API_NAME] The length of tag_key:{tag_key} and user_id: {user_id} can not be greater than 255"
        )
        NO_CAMPAIGN_FOUND = (
            "({file}): [API_NAME] No campaign found for goal_identifier:{goal_identifier}. Please verify from VWO app."
        )



# Copyright 2019-2021 Wingify Software Pvt. Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Utility module for manipulating VWO campaigns """

from __future__ import division
import math
from ..constants import constants
from ..enums.log_message_enum import LogMessageEnum
from ..enums.file_name_enum import FileNameEnum
from ..enums.log_level_enum import LogLevelEnum
from ..logger import VWOLogger

FILE = FileNameEnum.Helpers.CampaignUtil


def get_campaign(settings_file, campaign_key):
    """ Finds and Returns campaign from given campaign_key.

    Args:
        settings_file (dict): Settings file for the project
        campaign_key (string): Campaign identifier key

    Returns:
        dict: Campaign object
    """

    for campaign in settings_file.get("campaigns"):
        if campaign.get("key") == campaign_key and campaign.get("status") == constants.STATUS_RUNNING:
            return campaign
    VWOLogger.getInstance().log(
        LogLevelEnum.ERROR,
        LogMessageEnum.ERROR_MESSAGES.CAMPAIGN_NOT_RUNNING.format(file=FILE, campaign_key=campaign_key),
    )
    return None


def get_campaigns(settings_file, campaign_keys):
    """ Finds and Returns campaign from given campaign_keys.

    Args:
        settings_file (dict): Settings file for the project
        campaign_keys (list): List of Campaign identifier keys

    Returns:
        dict: Dictionary of campaign object mapped to the campaign key
    """
    campaign_key_map = {}
    for campaign in settings_file.get("campaigns"):
        campaign_key_map[campaign.get("key")] = campaign
    found_campaigns = {}
    for campaign_key in campaign_keys:
        if (
            campaign_key_map.get(campaign_key)
            and campaign_key_map.get(campaign_key).get("status") == constants.STATUS_RUNNING
        ):
            found_campaigns[campaign_key] = campaign_key_map.get(campaign_key)
        else:
            VWOLogger.getInstance().log(
                LogLevelEnum.ERROR,
                LogMessageEnum.ERROR_MESSAGES.CAMPAIGN_NOT_RUNNING.format(file=FILE, campaign_key=campaign_key),
            )
    return found_campaigns


def set_variation_allocation(campaign):
    """ Sets variation allocation range in the provided campaign.

    Args:
        campaign (dict): Campaign object
    """
    variation_allocations_ranges = get_variation_allocation_ranges(campaign.get("variations"))
    set_variation_allocation_from_ranges(campaign.get("variations"), variation_allocations_ranges)
    for variation in campaign.get("variations"):
        VWOLogger.getInstance().log(
            LogLevelEnum.INFO,
            LogMessageEnum.INFO_MESSAGES.VARIATION_RANGE_ALLOCATION.format(
                file=FILE,
                campaign_key=campaign.get("key"),
                variation_name=variation.get("name"),
                variation_weight=variation.get("weight"),
                start=variation.get("start_variation_allocation"),
                end=variation.get("end_variation_allocation"),
            ),
        )


def set_variation_allocation_from_ranges(variations, variation_allocations_ranges):
    """Sets variations allocation ranges on each variation as start_variation_allocation and
    end_variation_allocation.

    Args:
        variations(list): List of variation objects
        variation_allocation_ranges(list): List of tuples having starting and ending variation
        allocation range for each variation
    """
    for i, variation in enumerate(variations):
        variation.update(
            start_variation_allocation=variation_allocations_ranges[i][0],
            end_variation_allocation=variation_allocations_ranges[i][1],
        )


def get_variation_allocation_ranges(variations):
    """ Returns a list of variation allocation ranges.

    Args:
        variations (list): List of variation(dict object)

    Returns:
        list(tuple): list of tuple(start_range, end_range) for each
        variation
    """
    current_allocation = 0
    variation_allocation_ranges = []
    for variation in variations:
        step_factor = _get_variation_bucketing_range(variation.get("weight"))
        if step_factor:
            start_range = current_allocation + 1
            end_range = current_allocation + step_factor
            variation_allocation_ranges.append((start_range, end_range))
            current_allocation += step_factor
        else:
            variation_allocation_ranges.append((-1, -1))
    return variation_allocation_ranges


def _get_variation_bucketing_range(weight):
    """ Returns the bucket size of variation.

    Args:
        weight (int|float): weight of variation

    Returns:
        int: Bucket start range of Variation
    """

    if weight is None or weight == 0:
        return 0
    start_range = int(math.ceil(weight * 100))
    return min(start_range, constants.MAX_TRAFFIC_VALUE)


def get_campaigns_with_goal_id(campaigns, goal_identifer):
    """ Returns campaigns having the same goal_identifier passed in the args
    from the campaigns list

    Args:
        campaigns (list): List of campaign objects
        gaol_identifier (str): Global goal identifier

    Returns:
        tuple (campaign_goal_list, campaigns_without_goal): campaign_goal_list is a tuple
        of (campaign, campaign_goal)
    """
    campaign_goal_list = []
    campaigns_without_goal = []
    for campaign in campaigns:
        campaign_goal = get_campaign_goal(campaign, goal_identifer)
        if campaign_goal:
            campaign_goal_list.append((campaign, campaign_goal))
        else:
            campaigns_without_goal.append(campaign)
    return campaign_goal_list, campaigns_without_goal


def get_campaign_goal(campaign, goal_identifier):
    """ Returns goal from given campaign and Goal_identifier.

    Args:
        campaign (dict): The running campaign
        goal_identifier (string): Goal identifier

    Returns:
        dict: Goal corresponding to goal_identifer in respective campaign
    """

    if not campaign or not goal_identifier:
        return None
    for goal in campaign.get("goals"):
        if goal.get("identifier") == goal_identifier:
            return goal
    return None


def get_campaign_variation(campaign, variation_name):
    """ Returns variation from given campaign and variation_name.

    Args:
        campaign (dict): The running campaign
        variation_name (string): Variation identifier

    Returns:
        dict: Variation corresponding to variation_name in respective campaign
    """

    if not campaign or not variation_name:
        return None
    for variation in campaign.get("variations"):
        if variation.get("name") == variation_name:
            return variation
    return None


def get_variable(variables, variable_key):
    """ Returns variable from given variables list.

    Args:
        variables (list): List of variables, whether in campaigns or
        inside variation
        variable_key (string): Variable identifier

    Returns:
        dict: Variable corresponding to variable_key in given variable list
    """
    for variable in variables:
        if variable.get("key") == variable_key:
            return variable
    return None


def get_control_variation(campaign):
    """ Returns control variation from a given campaign

    Args:
        campaign (dict): Running campaign
    Returns:
        variation (dict): Control variation from the campaign, ie having id = 1
    """

    for variation in campaign.get("variations"):
        if int(variation.get("id")) == 1:
            return variation
    return None


def scale_variations(variations):
    """ It extracts the weights from all the variations inside the campaign
    and scales them so that the total sum of eligible variations' weights become 100%

    Args:
        variations(list): list of variations(dict object) having weight as a property
    """
    weight_sum = sum(variation.get("weight") for variation in variations)
    if weight_sum == 0:
        normalized_weight = 100 / len(variations)
        for variation in variations:
            variation["weight"] = normalized_weight
    else:
        for variation in variations:
            variation["weight"] = (variation["weight"] / weight_sum) * 100



# Copyright 2019-2021 Wingify Software Pvt. Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..http.connection import Connection
from ..enums.log_message_enum import LogMessageEnum
from ..enums.file_name_enum import FileNameEnum
from ..enums.log_level_enum import LogLevelEnum
from ..logger import VWOLogger

FILE = FileNameEnum.Event.EventDispatcher


class EventDispatcher(object):
    """ Class having request making/event dispatching capabilities to our servers"""

    def __init__(self, is_development_mode=False):
        """ Initialize the dispatcher with logger

        Args:
            is_development_mode: To specify whether the request
            to our server should be made or not.
        """
        self.logger = VWOLogger.getInstance()
        self.is_development_mode = is_development_mode
        self.connection = Connection()

    def dispatch(self, impression):
        """ This method checks for development mode, if it is False then it sends the impression
        to our servers using a vwo.http.connection.Connection object, else return True without
        sending the impression.

        Args:
            impression (dict): Dictionary object containing the information of the impression

        Returns:
            bool: True if impression is successfully received by our servers, else false
        """
        url = impression.pop("url")
        if self.is_development_mode:
            result = True
        else:
            resp = self.connection.get(url, params=impression)
            result = resp.get("status_code") == 200

        if result is True:
            self.logger.log(
                LogLevelEnum.INFO, LogMessageEnum.INFO_MESSAGES.IMPRESSION_SUCCESS.format(file=FILE, end_point=url)
            )
            return True
        else:
            self.logger.log(
                LogLevelEnum.ERROR, LogMessageEnum.ERROR_MESSAGES.IMPRESSION_FAILED.format(file=FILE, end_point=url)
            )
            return False



# Copyright 2019-2021 Wingify Software Pvt. Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import pkg_resources
import logging

SDK_NAME = "vwo-python-sdk"
SDK_VERSION = pkg_resources.require(SDK_NAME)[0].version
API_VERSION = 1
PLATFORM = "server"
SEED_VALUE = 1
MAX_TRAFFIC_PERCENT = 100
MAX_TRAFFIC_VALUE = 10000
STATUS_RUNNING = "RUNNING"
LIBRARY_PATH = os.path.normpath(os.getcwd() + os.sep + os.pardir)
HTTP_PROTOCOL = "http://"
HTTPS_PROTOCOL = "https://"


class ENDPOINTS:
    BASE_URL = "dev.visualwebsiteoptimizer.com"
    ACCOUNT_SETTINGS = "/server-side/settings"
    TRACK_USER = "/server-side/track-user"
    TRACK_GOAL = "/server-side/track-goal"
    PUSH = "/server-side/push"


class EVENTS:
    TRACK_USER = "track-user"
    TRACK_GOAL = "track-goal"


class DATA_TYPE:
    NUMBER = "number"
    STRING = "string"
    FUNCTION = "function"
    BOOLEAN = "boolean"


class API_METHODS:
    ACTIVATE = "activate"
    GET_VARIATION_NAME = "get_variation_name"
    TRACK = "track"
    IS_FEATURE_ENABLED = "is_feature_enabled"
    GET_FEATURE_VARIABLE_VALUE = "get_feature_variable_value"
    PUSH = "push"


class GOAL_TYPES:
    REVENUE = "REVENUE_TRACKING"
    CUSTOM = "CUSTOM_GOAL"
    ALL = "ALL"


class CAMPAIGN_TYPES:
    VISUAL_AB = "VISUAL_AB"
    FEATURE_TEST = "FEATURE_TEST"
    FEATURE_ROLLOUT = "FEATURE_ROLLOUT"


class VARIABLE_TYPES:
    STRING = "string"
    INTEGER = "integer"
    DOUBLE = "double"
    BOOLEAN = "boolean"


class PUSH_API:
    TAG_VALUE_LENGTH = 255
    TAG_KEY_LENGTH = 255


PY_VARIABLE_TYPES = {"string": str, "integer": int, "double": float, "boolean": bool}


# For exposing log levels to vwo_instance
class LOG_LEVELS:
    DEBUG = logging.DEBUG
    INFO = logging.INFO
    WARNING = logging.WARNING
    ERROR = logging.ERROR



# Copyright 2019-2021 Wingify Software Pvt. Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Generating UUID required for sending impression to server """

import uuid
from ..enums.log_message_enum import LogMessageEnum
from ..enums.file_name_enum import FileNameEnum
from ..enums.log_level_enum import LogLevelEnum
from ..logger import VWOLogger

VWO_NAMESPACE = uuid.uuid5(uuid.NAMESPACE_URL, "https://vwo.com")
FILE = FileNameEnum.Helpers.UuidUtil


def generate_for(user_id, account_id):
    """ Generates desired UUID

    Args:
        user_id (int|string): User identifier
        account_id (int|string): Account identifier

    Returns:
        int : Desired Uuid
    """

    user_id = str(user_id)
    account_id = str(account_id)
    user_id_namespace = generate(VWO_NAMESPACE, account_id)
    uuid_for_account_user_id = generate(user_id_namespace, user_id)

    desired_uuid = str(uuid_for_account_user_id).replace("-", "").upper()

    VWOLogger.getInstance().log(
        LogLevelEnum.DEBUG,
        LogMessageEnum.DEBUG_MESSAGES.UUID_FOR_USER.format(
            file=FILE, user_id=user_id, account_id=account_id, desired_uuid=desired_uuid
        ),
    )
    return desired_uuid


def generate(namespace, name):
    """ Generated uuid from namespace and name, uses uuid5 from python uuid module

    Args:
        namespace (string): Namespace
        name (string): Name

    Returns:
        string|None: Uuid, none if any of the arguments is empty
    """

    if name and namespace:
        return uuid.uuid5(namespace, name)
    return None



# Copyright 2019-2021 Wingify Software Pvt. Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class UserStorage(object):
    """ UserStorage Class is used to store user-variation mapping.
    Override this class to implement your own functionality.
    SDK will ensure to use this while bucketing a user into a variation."""

    def get(self, user_id, campaign_key):
        """ To retrieve the stored variation for the user_id and
        campaign_key

        Args:
            user_id (str): User ID for which data needs to be retrieved.
            campaign_key (str): Campaign key to identify the campaign for
            which stored variation should be retrieved.

        Returns:
            user_data (dict): user-variation mapping
        """
        pass

    def set(self, user_data):
        """ To store the the user variation-mapping

        Args:
            user_data (dict): user-variation mapping
        """
        pass



# Copyright 2019-2021 Wingify Software Pvt. Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class FileNameEnum:
    """ Classobj encapsulating locations various files of the SDK """

    class Vwo:
        VWO_PATH = "vwo/"
        VWO = VWO_PATH + "vwo"

    class Api:
        API_PATH = "vwo/api/"
        Activate = API_PATH + "activate"
        GetVariationName = API_PATH + "get_variation_name"
        Track = API_PATH + "track"
        IsFeatureEnabled = API_PATH + "is_feature_enabled"
        GetFeatureVariableValue = API_PATH + "get_feature_variable_value"
        Push = API_PATH + "push"
        Launch = API_PATH + "launch"

    class Core:
        CORE_PATH = "vwo/core/"
        Bucketer = CORE_PATH + "bucketer"
        VariationDecider = CORE_PATH + "variation_decider"

    class Event:
        EVENT_PATH = "vwo/event/"
        EventDispatcher = EVENT_PATH + "event_dispatcher"

    class Helpers:
        HELPERS_PATH = "vwo/helpers/"
        CampaignUtil = HELPERS_PATH + "campaign_util"
        CustomDimensionsUtil = HELPERS_PATH + "custom_dimensions_util"
        FeatureUtil = HELPERS_PATH + "feature_util"
        GenericUtil = HELPERS_PATH + "generic_util"
        ImpressionUtil = HELPERS_PATH + "impression_util"
        UuidUtil = HELPERS_PATH + "uuid_util"
        ValidateUtil = HELPERS_PATH + "validate_util"

    class Logger:
        LOGGER_PATH = "vwo/logger/"
        LoggerManager = LOGGER_PATH + "logger_manager"

    class Services:
        SERVICES_PATH = "vwo/services/"
        SettingsFileManager = SERVICES_PATH + "settings_file_manager"
        SegmentEvaluator = SERVICES_PATH + "segment_evaluator"



# Copyright 2019-2021 Wingify Software Pvt. Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
from ..enums.log_message_enum import LogMessageEnum
from ..enums.file_name_enum import FileNameEnum
from ..enums.log_level_enum import LogLevelEnum
from ..enums.segments.result_status import ResultStatus
from ..helpers import validate_util, campaign_util
from ..logger import VWOLogger
from .bucketer import Bucketer
from ..services.segmentor import SegmentEvaluator
from ..constants import constants

FILE = FileNameEnum.Core.VariationDecider


class VariationDecider(object):
    """ Class responsible for deciding the variation for a visitor """

    def __init__(self, user_storage=None):
        """ Initializes VariationDecider with settings_file,
            UserStorage and logger.

        Args:
            settings_file (dict): settings_file consisting all the campaign
            related data
            user_storage: Class instance having the capability of
                get and set.
        """
        self.logger = VWOLogger.getInstance()
        self.bucketer = Bucketer()
        self.segment_evaluator = SegmentEvaluator()
        self.user_storage = user_storage

    def get_variation(self, user_id, campaign, **kwargs):
        """ Returns variation for the user for given campaign
            This method achieves the variation assignment in the following way:
            1. First get variation from UserStorage, if variation is found in user_storage_data,
                return from there
            2. Evaluates white listing users for each variation, and find a targeted variation.
            3. If no targeted variation is found, evaluate pre-segmentation result
            4. Evaluate percent traffic
            5. If user becomes part of campaign assign a variation.
            6. Store the variation found in the user_storage

        Args:
            user_id (string): the unique ID assigned to User
            campaign (dict): campaign in which user is participating
            custom_variables(dict): variables for pre-segmentation, pass it through **kwargs as
            custom_variables = {}
            variation_targeting_variables(dict): variables for variation targeting, pass it through **kwargs as
            variation_targeting_variables = {}

        Returns:
            variation (dict|None): Dict object containing the information regarding variation
            assigned else None
        """

        custom_variables = kwargs.get("custom_variables")
        variation_targeting_variables = kwargs.get("variation_targeting_variables")
        goal_data = kwargs.get("goal_data")
        # Evaluate whitelisting at first
        targeted_variation = self.find_targeted_variation(user_id, campaign, variation_targeting_variables)
        if targeted_variation:
            self.logger.log(
                LogLevelEnum.INFO,
                LogMessageEnum.INFO_MESSAGES.GOT_VARIATION_FOR_USER.format(
                    file=FILE,
                    variation_name=targeted_variation.get("name"),
                    user_id=user_id,
                    campaign_key=campaign.get("key"),
                    campaign_type=campaign.get("type"),
                ),
            )
            return targeted_variation

        # Try retrieving data from user_storage
        user_storage_data = self._get_user_storage_data(user_id, campaign.get("key"))
        if user_storage_data:
            # If being called by track, check for goals identified
            if goal_data:
                is_goal_tracked = self.identify_tracked_goal_from_user_storage(goal_data, user_storage_data)
                if is_goal_tracked:
                    self.logger.log(
                        LogLevelEnum.INFO,
                        LogMessageEnum.INFO_MESSAGES.GOAL_ALREADY_TRACKED.format(
                            file=FILE,
                            goal_identifier=goal_data.get("identifier"),
                            campaign_key=campaign.get("key"),
                            user_id=user_id,
                        ),
                    )
                    return None
            # Retreving variation from user_storage_data
            variation = self.get_variation_from_user_storage(user_id, campaign, user_storage_data)
            # Setting goal information if variation found and api is track
            if variation is not None:
                if goal_data:
                    self.update_goals_tracked_in_user_storage(goal_data, user_storage_data)
                return variation

        # Evaluate pre-segmentation and percent-traffic
        if self.evaluate_pre_segmentation(user_id, campaign, custom_variables) and self.is_user_part_of_campaign(
            user_id, campaign
        ):
            variation = self.bucketer.bucket_user_to_variation(user_id, campaign)
            new_user_storage_data = self._create_user_storage_data(
                user_id, campaign.get("key"), variation.get("name"), goal_data=kwargs.get("goal_data")
            )
            self._set_user_storage_data(new_user_storage_data)
            self.logger.log(
                LogLevelEnum.INFO,
                LogMessageEnum.INFO_MESSAGES.GOT_VARIATION_FOR_USER.format(
                    file=FILE,
                    variation_name=variation.get("name"),
                    user_id=user_id,
                    campaign_key=campaign.get("key"),
                    campaign_type=campaign.get("type"),
                ),
            )
            return variation

        # No variation
        self.logger.log(
            LogLevelEnum.INFO,
            LogMessageEnum.INFO_MESSAGES.USER_GOT_NO_VARIATION.format(
                file=FILE, user_id=user_id, campaign_key=campaign.get("key"), campaign_type=campaign.get("type")
            ),
        )
        return None

    def identify_tracked_goal_from_user_storage(self, goal_data, user_storage_data):
        """ Identifies whether the given goal has been already tracked or not.

        Args:
            goal_data (dict): goal related data
            user_storage_data (dict): data retrieved from user_storage

        Returns:
            bool: True if goal already tracked else False
        """
        goals_tracked = user_storage_data.get("goalIdentifiers")
        if (
            goals_tracked
            and (goal_data.get("identifier") in goals_tracked.split("_vwo_"))
            and not goal_data.get("should_track_returning_user")
        ):
            return True
        return False

    def update_goals_tracked_in_user_storage(self, goal_data, user_storage_data):
        """ Updates the goals tracked information stored inside the user_storage service

        Args:
            goal_data (dict): the goal related data
            user_storage_data (dict): new or retrieved user_storage_data
        """
        goals_tracked = user_storage_data.get("goalIdentifiers")
        if goals_tracked:
            updated_goals_tracked = goals_tracked + "_vwo_" + goal_data.get("identifier")
        else:
            updated_goals_tracked = goal_data.get("identifier")
        user_storage_data["goalIdentifiers"] = updated_goals_tracked
        self._set_user_storage_data(user_storage_data)

    def get_variation_from_user_storage(self, user_id, campaign, user_storage_data):
        """ Tries retrieving variation from user_storage

        Args:
            user_id(string): unique user identifier
            campaign(dict): campaign for which the variation is to be retrieved

        Returns:
            variation (dict|None): Dict object containing the information regarding variation
            assigned else None
        """
        if validate_util.is_valid_dict(user_storage_data):
            if user_storage_data.get("campaignKey") == campaign.get("key"):
                variation_name = user_storage_data.get("variationName")
                if validate_util.is_valid_string(variation_name):
                    variation = campaign_util.get_campaign_variation(campaign, variation_name)
                    if variation:
                        self.logger.log(
                            LogLevelEnum.INFO,
                            LogMessageEnum.INFO_MESSAGES.GOT_STORED_VARIATION.format(
                                file=FILE,
                                campaign_key=campaign.get("key"),
                                user_id=user_id,
                                variation_name=variation.get("name"),
                            ),
                        )
                        return variation
        self.logger.log(
            LogLevelEnum.DEBUG,
            LogMessageEnum.DEBUG_MESSAGES.NO_STORED_VARIATION.format(
                file=FILE, campaign_key=campaign.get("key"), user_id=user_id
            ),
        )
        return None

    def find_targeted_variation(self, user_id, campaign, variation_targeting_variables):
        """ Identifies and retrives if there exists any targeted variation in the given campaign
        for given user_id

        Args:
            user_id(string): unique user identifier
            campaign(dict): campaign for which the variation is to be retrieved
            variation_targeting_variables(dict): variables for finding targeted variation

        Returns:
            targeted_variation (dict|None): Dict object containing the information regarding forced
            variation assigned else None
        """
        if campaign.get("isForcedVariationEnabled") is not True:
            self.logger.log(
                LogLevelEnum.DEBUG,
                LogMessageEnum.DEBUG_MESSAGES.WHITELISTING_SKIPPED.format(
                    file=FILE, user_id=user_id, campaign_key=campaign.get("key")
                ),
            )
            return None
        else:
            white_listed_variations_list = self._get_white_listed_variations_list(
                user_id, campaign, variation_targeting_variables
            )
            white_listed_variations_len = len(white_listed_variations_list)
            if white_listed_variations_len == 0:
                targeted_variation = None
            elif white_listed_variations_len == 1:
                targeted_variation = white_listed_variations_list[0]
            else:
                # Scale the traffic percent of each variation
                campaign_util.scale_variations(white_listed_variations_list)
                # Allocate new range
                variation_allocations = campaign_util.get_variation_allocation_ranges(white_listed_variations_list)
                campaign_util.set_variation_allocation_from_ranges(white_listed_variations_list, variation_allocations)
                # Now retrieve the variation from the modified_campaign_for_whitelisting
                bucket_value = self.bucketer.get_bucket_value_for_user(user_id, constants.MAX_TRAFFIC_VALUE)
                targeted_variation = self.bucketer.get_variation(white_listed_variations_list, bucket_value)
            variation_status = "and variation {variation_name} is assigned"
            self.logger.log(
                LogLevelEnum.INFO,
                LogMessageEnum.INFO_MESSAGES.SEGMENTATION_STATUS.format(
                    file=FILE,
                    campaign_key=campaign.get("key"),
                    user_id=user_id,
                    variables=variation_targeting_variables,
                    variation_status=variation_status.format(
                        variation_name=targeted_variation.get("name") if targeted_variation else None
                    ),
                    segmentation_type="whitelisting",
                    status=ResultStatus.PASSED if targeted_variation is not None else ResultStatus.FAILED,
                ),
            )
            return targeted_variation

    def evaluate_pre_segmentation(self, user_id, campaign, custom_variables):
        """ Evaluates segmentation for the user_id against the segments found inside
        the campaign.

        Args:
            user_id(string): unique user identifier
            campaign(dict): running campaign for which the segments is to be evaluated
            custom_variables(dict): variables for segmentation

        Returns:
            bool: True if user passes segmentation, else False
        """
        segments = campaign.get("segments")
        if not validate_util.is_valid_value(segments):
            result = True
            self.logger.log(
                LogLevelEnum.DEBUG,
                LogMessageEnum.DEBUG_MESSAGES.SEGMENTATION_SKIPPED.format(
                    file=FILE,
                    user_id=user_id,
                    campaign_key=campaign.get("key"),
                    variables=custom_variables,
                    variation_status="",
                ),
            )
        else:
            if not validate_util.is_valid_value(custom_variables):
                self.logger.log(
                    LogLevelEnum.DEBUG,
                    LogMessageEnum.DEBUG_MESSAGES.NO_VARIABLES.format(
                        file=FILE,
                        user_id=user_id,
                        campaign_key=campaign.get("key"),
                        segmentation_type="pre_segmentation",
                    ),
                )
                custom_variables = {}
            try:
                result = self.segment_evaluator.evaluate(segments, custom_variables)
                self.logger.log(
                    LogLevelEnum.INFO,
                    LogMessageEnum.INFO_MESSAGES.SEGMENTATION_STATUS.format(
                        file=FILE,
                        user_id=user_id,
                        campaign_key=campaign.get("key"),
                        variables=custom_variables,
                        variation_status="",
                        segmentation_type="pre_segmentation",
                        status=ResultStatus.PASSED if result else ResultStatus.FAILED,
                    ),
                )
            except Exception as e:
                result = False
                self.logger.log(
                    LogLevelEnum.ERROR,
                    LogMessageEnum.ERROR_MESSAGES.SEGMENTATION_ERROR.format(
                        file=FILE,
                        user_id=user_id,
                        campaign_key=campaign.get("key"),
                        variables=custom_variables,
                        variation_status="",
                        error_message=e,
                    ),
                )
        return result

    def is_user_part_of_campaign(self, user_id, campaign):
        """ Evaluates whether the user should become part of campaign
        or not

        Args:
            user_id (string): the unique ID assigned to User
            campaign (dict): campaign in which user is participating

        Returns:
            bool: True if user should become part of campaign, else False
        """

        if self.bucketer.is_user_part_of_campaign(user_id, campaign):
            return True
        else:
            # not part of campaign
            self.logger.log(
                LogLevelEnum.DEBUG,
                LogMessageEnum.DEBUG_MESSAGES.USER_NOT_PART_OF_CAMPAIGN.format(
                    file=FILE,
                    user_id=user_id,
                    campaign_key=campaign.get("key"),
                    method="is_user_part_of_campaign",
                    campaign_type=campaign.get("type"),
                ),
            )
            return False

    # Private helper methods

    def _get_white_listed_variations_list(self, user_id, campaign, variation_targeting_variables):
        """ Identifies all forced variations which are targeted by variation_targeting_variables

        Args:
            user_id(string): unique user identifier
            campaign(dict): campaign for which the targeted variation(s) is to be retrieved
            variation_targeting_variables(dict): variables for variation targeting

        Returns:
            targeted_variation (list): List of targeted variation objects
        """
        if not validate_util.is_valid_value(variation_targeting_variables):
            self.logger.log(
                LogLevelEnum.DEBUG,
                LogMessageEnum.DEBUG_MESSAGES.NO_VARIABLES.format(
                    file=FILE, user_id=user_id, campaign_key=campaign.get("key"), segmentation_type="whitelisting"
                ),
            )
            variation_targeting_variables = {}

        variation_targeting_variables.update(_vwo_user_id=user_id)
        white_listed_variations_list = []

        for variation in campaign.get("variations"):
            if not validate_util.is_valid_value(variation.get("segments")):
                result = False
                self.logger.log(
                    LogLevelEnum.DEBUG,
                    LogMessageEnum.DEBUG_MESSAGES.SEGMENTATION_SKIPPED.format(
                        file=FILE,
                        user_id=user_id,
                        variation_name=variation.get("name"),
                        campaign_key=campaign.get("key"),
                        variation_status="for variation %s" % variation.get("name"),
                    ),
                )
            else:
                try:
                    result = self.segment_evaluator.evaluate(variation.get("segments"), variation_targeting_variables)
                    self.logger.log(
                        LogLevelEnum.DEBUG,
                        LogMessageEnum.DEBUG_MESSAGES.SEGMENTATION_STATUS.format(
                            file=FILE,
                            user_id=user_id,
                            status="passed" if result else "failed",
                            variables=variation_targeting_variables,
                            variation_status="for variation %s" % variation.get("name"),
                            campaign_key=campaign.get("key"),
                            segmentation_type="white_listing",
                        ),
                    )
                except Exception as e:
                    result = False
                    self.logger.log(
                        LogLevelEnum.ERROR,
                        LogMessageEnum.ERROR_MESSAGES.SEGMENTATION_ERROR.format(
                            file=FILE,
                            user_id=user_id,
                            variables=variation_targeting_variables,
                            campaign_key=campaign.get("key"),
                            variation_status=" for variation %s" % variation.get("name"),
                            error_message=e,
                        ),
                    )
            if result:
                white_listed_variations_list.append(copy.deepcopy(variation))
        return white_listed_variations_list

    def _create_user_storage_data(self, user_id, campaign_key, variation_name, **kwargs):
        """ Creates a user_storage_data object to be set by user_storage service implemented by user

        Args:
            user_id (str): unique user ID
            campaign_key (str): campaign key
            variation_name (str): winner of the campaign

        Keyword Args:
            goal_data (dict): goal related data if track api being called

        Returns:
            dict: a user_storage_data object with specified properties
        """
        new_user_storage_data = {"userId": user_id, "campaignKey": campaign_key, "variationName": variation_name}
        # For track only, to store the goals tracked
        if kwargs.get("goal_data"):
            new_user_storage_data["goalIdentifiers"] = kwargs.get("goal_data").get("identifier")
        return new_user_storage_data

    def _get_user_storage_data(self, user_id, campaign_key):
        """ Get the UserStorageData after looking up into get method
        being provided via UserStorage service

        Args:
            user_id (string): Unique user identifier
            campaign_key (string): Unique campaign identifier
        Returns:
            dict: user_storage_data data
        """

        if not self.user_storage:
            self.logger.log(LogLevelEnum.DEBUG, LogMessageEnum.DEBUG_MESSAGES.NO_USER_STORAGE_GET.format(file=FILE))
            return False
        try:
            user_storage_data = self.user_storage.get(user_id, campaign_key)
            self.logger.log(
                LogLevelEnum.INFO,
                LogMessageEnum.INFO_MESSAGES.LOOKING_UP_USER_STORAGE.format(file=FILE, user_id=user_id),
            )
            return copy.deepcopy(user_storage_data)
        except Exception:
            self.logger.log(
                LogLevelEnum.ERROR,
                LogMessageEnum.ERROR_MESSAGES.LOOK_UP_USER_STORAGE_FAILED.format(file=FILE, user_id=user_id),
            )
            return False

    def _set_user_storage_data(self, user_storage_data):
        """ If UserStorage is provided and variation was found,
        set the assigned variation in UserStorage.
            It creates bucket and then stores.

        Args:
            user_id (string): Unique user identifier
            campaign_key (string): Unique campaign identifier
            variation_name (string): variation identifier

        Returns:
            bool: true if found otherwise false
        """

        if not self.user_storage:
            self.logger.log(LogLevelEnum.DEBUG, LogMessageEnum.DEBUG_MESSAGES.NO_USER_STORAGE_SET.format(file=FILE))
            return False

        try:
            self.user_storage.set(user_storage_data)
            self.logger.log(
                LogLevelEnum.INFO,
                LogMessageEnum.INFO_MESSAGES.SAVING_DATA_USER_STORAGE_STATUS.format(
                    file=FILE, user_id=user_storage_data.get("userId"), status="successful"
                ),
            )
            return True
        except Exception as e:
            self.logger.log(
                LogLevelEnum.ERROR,
                LogMessageEnum.ERROR_MESSAGES.SET_USER_STORAGE_FAILED.format(
                    file=FILE, user_id=user_storage_data.get("userId"), error_message=e
                ),
            )
            return False



# Copyright 2019-2021 Wingify Software Pvt. Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..helpers import custom_dimensions_util
from ..constants import constants
from ..constants.constants import API_METHODS
from ..helpers import validate_util
from ..enums.log_message_enum import LogMessageEnum
from ..enums.file_name_enum import FileNameEnum
from ..enums.log_level_enum import LogLevelEnum

FILE = FileNameEnum.Api.Push


def _push(vwo_instance, tag_key, tag_value, user_id):
    """ This API method: Makes a call to our server to store the tag_values

    1. Validates the arguments being passed
    2. Send a call to our server

    Args:
        tag_key (string): key name of the tag
        tag_value (string): value of the tag
        user_id (string): ID of the user for which value should be stored

    Returns:
        bool : True if call is made successfully, else False
    """

    vwo_instance.logger.set_api(API_METHODS.PUSH)
    if (
        not validate_util.is_valid_string(tag_key)
        or not validate_util.is_valid_string(tag_value)
        or not validate_util.is_valid_string(user_id)
    ):
        vwo_instance.logger.log(
            LogLevelEnum.ERROR, LogMessageEnum.ERROR_MESSAGES.PUSH_API_INVALID_PARAMS.format(file=FILE)
        )
        return False

    if len(tag_key) > constants.PUSH_API.TAG_KEY_LENGTH:
        vwo_instance.logger.log(
            LogLevelEnum.ERROR,
            LogMessageEnum.ERROR_MESSAGES.TAG_KEY_LENGTH_EXCEEDED.format(file=FILE, user_id=user_id, tag_key=tag_key),
        )
        return False

    if len(tag_value) > constants.PUSH_API.TAG_VALUE_LENGTH:
        vwo_instance.logger.log(
            LogLevelEnum.ERROR,
            LogMessageEnum.ERROR_MESSAGES.TAG_VALUE_LENGTH_EXCEEDED.format(
                file=FILE, user_id=user_id, tag_value=tag_value
            ),
        )
        return False

    impression = custom_dimensions_util.get_url_params(vwo_instance.settings_file, tag_key, tag_value, user_id)

    vwo_instance.event_dispatcher.dispatch(impression)

    vwo_instance.logger.log(
        LogLevelEnum.INFO,
        LogMessageEnum.INFO_MESSAGES.MAIN_KEYS_FOR_PUSH_API.format(
            file=FILE,
            u=impression.get("u"),
            user_id=impression.get("uId"),
            account_id=impression.get("account_id"),
            tags=impression.get("tags"),
        ),
    )
    return True



# Copyright 2019-2021 Wingify Software Pvt. Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from ..enums.log_message_enum import LogMessageEnum
from ..enums.file_name_enum import FileNameEnum
from ..enums.log_level_enum import LogLevelEnum
from ..helpers import campaign_util
from ..logger import VWOLogger

FILE = FileNameEnum.Services.SettingsFileManager


class SettingsFileManager(object):
    """ VWO settings_file manager """

    def __init__(self, settings_file):
        """ Init method to load and set vwo object with settings_file data.

        Args:
            settings_file (dict): Dict object
                representing the vwo settings_file.
        """
        self.settings_file = json.loads(settings_file)
        self.logger = VWOLogger.getInstance()
        self.process_settings_file()

    # PUBLIC METHODS
    def process_settings_file(self):
        """ Processes the settings_file, assigns variation allocation range """

        settings_file = self.settings_file
        for campaign in settings_file.get("campaigns"):
            campaign_util.set_variation_allocation(campaign)
        self.logger.log(LogLevelEnum.DEBUG, LogMessageEnum.DEBUG_MESSAGES.SETTINGS_FILE_PROCESSED.format(file=FILE))

    def get_settings_file(self):
        """ Retrieves settings file """

        return self.settings_file



# Copyright 2019-2021 Wingify Software Pvt. Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import math
import re
import sys

from ..enums.segments import OperandValueTypesName, OperandValuesBooleanTypes, OperandValueTypes

# This regex returns array of tuples of (operand_type, operand_value)
# from the string of type "operand_type(operand_value)"
GROUPING_PATTERN = re.compile("^(.+?)\((.*)\)")  # noqa: W605

# This regex returns array of tuples of (starting_star, operand_value, ending_star)
# from the string of type "*operand_value*"
WILDCARD_PATTERN = re.compile("(^\*|^)(.+?)(\*$|$)")  # noqa: W605


def convert_to_true_types(operator_value, custom_variables_value):
    """ Extracts true values represented in the args, and returns
    stringified value of it

    Args:
        operator_value(str): operand/dsl leaf value
        custom_variables_value(int|str|float): custom_variables value

    Returns:
        (str, str): tuple of str value of operator_value, custom_variables_value converted
        into their true types
    """

    # To deal with unicode type in python 2, if the type is unicode
    # encode the values
    if sys.version_info[0] < 3:
        if type(operator_value) is unicode:
            operator_value = operator_value.encode("utf-8")
        if type(custom_variables_value) is unicode:
            custom_variables_value = custom_variables_value.encode("utf-8")

    # This is atomic, either both values will be processed or none
    try:
        true_type_operator_value = float(operator_value)
        true_type_custom_variables_value = float(custom_variables_value)
    except Exception:
        return operator_value, custom_variables_value

    # Now both are float, So, convert them independently to int type
    # if they are int rather than floats
    if true_type_operator_value == math.floor(true_type_operator_value):
        true_type_operator_value = int(true_type_operator_value)
    if true_type_custom_variables_value == math.floor(true_type_custom_variables_value):
        true_type_custom_variables_value = int(true_type_custom_variables_value)

    # Convert them back to string and return
    return str(true_type_operator_value), str(true_type_custom_variables_value)


def separate_operand(operand):
    """ Extract the operand_type, ie. lower, wildcard, regex or equals

    Args:
        operand(str): string value from leaf_node of dsl

    Returns:
        (str, str): tuple of operand value and operand type
    """
    groups = GROUPING_PATTERN.findall(operand)
    if groups:
        return groups[0]
    return (OperandValueTypesName.EQUALS, operand)


def process_custom_variables_value(custom_variables_value):
    """ Processes the value from the custom_variables_variables

    Args:
        custom_variables_value(str|int|float|bool|None): the custom_variables_value provided
        inside custom_variables

    Returns:
        str(custom_variables_value): stringified value of processed custom_variables_value
    """
    if custom_variables_value is None:
        custom_variables_value = ""

    if type(custom_variables_value) is bool:
        if custom_variables_value:
            custom_variables_value = OperandValuesBooleanTypes.TRUE
        else:
            custom_variables_value = OperandValuesBooleanTypes.FALSE

    return str(custom_variables_value)


def process_operand_value(operand):
    """ Extracts operand_type and operand_value from the leaf_node/operand

    Args:
        operand(str): string value from the leaf_node

    Returns:
        (operand_type, operand_value): tuple of defined operand_types and
        operand_value
    """
    # separate the operand type and value inside the bracket
    operand_type_name, operand_value = separate_operand(operand)

    # Enum the operand type, here lower, regex, and equals will be identified
    operand_type = getattr(OperandValueTypes, operand_type_name, None)

    # In case of wildcard, the operand type is further divided into contains, startswith and endswith
    if operand_type_name == OperandValueTypesName.WILDCARD:
        starting_star, operand_value, ending_star = WILDCARD_PATTERN.findall(operand_value)[0]
        if starting_star and ending_star:
            operand_type = OperandValueTypes.contains
        elif starting_star:
            operand_type = OperandValueTypes.startswith
        elif ending_star:
            operand_type = OperandValueTypes.endswith
        else:
            operand_type = OperandValueTypes.equals

    # In case there is an abnormal patter, it would have passed all the above if cases, which means it
    # should be equals, so set the whole operand as operand value and operand type as equals
    if operand_type is None:
        operand_type, operand_value = OperandValueTypes.equals, operand

    return operand_type, operand_value



# Copyright 2019-2021 Wingify Software Pvt. Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..enums.file_name_enum import FileNameEnum
from ..enums.log_level_enum import LogLevelEnum
from ..enums.log_message_enum import LogMessageEnum
from ..helpers import validate_util
from ..vwo import VWO
from ..logger import VWOLogger

FILE = FileNameEnum.Api.Launch


def launch(settings_file, logger=None, user_storage=None, is_development_mode=False, **kwargs):
    """ Launch api is the gateway to the SDK, it accepts all the launch parameters and
    initializes required services for the SDK, like logger, event_dispatcher, etc, and
    returns a VWO instance.

    Args:
        settings_file (dict): settings_file consisting all the campaign related data
        logger (object): an object capable of logging events happening inside the SDK
        user_storage (object): a storage service object capable of doing get and set on
        SDK provide data
        is_development_mode (bool): should the SDK be initialized in development mode,
        it toggles the event_dispatcher to off

    Keyword Args:
        log_level (vwo.LOG_LEVELS): a log_level with which SDK should be initialized.
        Default value is vwo.LOG_LEVELS.ERROR
        goal_type_to_track (vwo.GOAL_TYPES): which goal type to track when using track
        api. Default value is vwo.GOAL_TYPES.ALL
        should_track_returning_user (bool): should returning user be tracked again.
        Default value is False

    Returns:
        VWO object: Successfully creates and returns a VWO object with passed params
        if all the params are valid else None
    """
    VWOLogger.clearExistingLoggerInstance()

    log_level = kwargs.get("log_level")
    goal_type_to_track = kwargs.get("goal_type_to_track")
    should_track_returning_user = kwargs.get("should_track_returning_user")

    invalid_log_level = False
    if log_level and not validate_util.is_valid_log_level(log_level):
        log_level = None
        invalid_log_level = True

    invalid_logger = False
    if logger and not validate_util.is_valid_service(logger, "logger"):
        logger = None
        invalid_logger = True
    logger = VWOLogger.getInstance(log_level=log_level, logger=logger)

    if (
        invalid_log_level
        or invalid_logger
        or (user_storage and not validate_util.is_valid_service(user_storage, "user_storage"))
        or not validate_util.is_valid_settings_file(settings_file)
        or (is_development_mode and type(is_development_mode) is not bool)
        or (goal_type_to_track and not validate_util.is_valid_goal_type(goal_type_to_track))
        or (should_track_returning_user and type(should_track_returning_user) is not bool)
    ):
        logger.log(LogLevelEnum.ERROR, LogMessageEnum.ERROR_MESSAGES.LAUNCH_API_INVALID_PARAMS.format(file=FILE))
        return None
    else:
        logger.log(LogLevelEnum.DEBUG, LogMessageEnum.DEBUG_MESSAGES.VALID_CONFIGURATION.format(file=FILE))
        return VWO(settings_file, user_storage, is_development_mode, goal_type_to_track, should_track_returning_user)



# Copyright 2019-2021 Wingify Software Pvt. Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Generic utility for helper math and random functions """

import random
from datetime import datetime
import time
from ..logger import VWOLogger
from ..enums.log_level_enum import LogLevelEnum
from ..enums.log_message_enum import LogMessageEnum

logger = VWOLogger.getInstance()


def get_random_number():
    """ Returns a random number

    Returns:
        float: A random number
    """
    return random.random()


def get_current_unix_timestamp():
    """ Returns current unix timestamp

    Returns:
        int: Current unix timestamp
    """
    return int(time.mktime(datetime.now().timetuple()))


def get_key_value(var):
    """ Returns first key value pair of the given dict. Use this only when the dict has one
    key-value pair as python2 and python3 sort dicts in different ways.

    Returns:
        var, var: Tuple of key value pair
    """
    key = next(iter(var))
    value = var[key]
    return key, value


def get_attribute_values(obj):
    """ Returns list of attribute values

    Args:
    obj(class): class obj

    Returns:
    attributes(dict): list of attribute values
    """
    return [value for key, value in obj.__dict__.items() if not key.startswith("__")]


def safe_method(method, fail_return_value, FILE):
    """ A decoratore to wrap API methods in fail safe manner

    Args:
        method (method): the API method which should be wrapped
        fail_return_value (value): a value which should be returned when the method passed
        fails
        FILE (str): path of the method for logging purpose

    Returns:
        method : a failsafe API method
    """

    def inner_method(*args, **kwargs):
        try:
            return method(*args, **kwargs)
        except Exception as e:
            logger.log(LogLevelEnum.ERROR, LogMessageEnum.ERROR_MESSAGES.API_NOT_WORKING.format(file=FILE, exception=e))
            return fail_return_value

    return inner_method



# Copyright 2019-2021 Wingify Software Pvt. Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# flake8: noqa

import os

from setuptools import setup, find_packages
from setuptools.command.develop import develop
from setuptools import Command

import subprocess

current_directory = os.path.join(os.path.dirname(__file__))

with open(os.path.join(current_directory, "requirements.txt")) as f:
    REQUIREMENTS = f.read().splitlines()

with open("README.md", "r") as f:
    long_description = f.read()


class DocCheckCommand(Command):
    description = "Doc Check"
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        subprocess.call("python3 ./scripts/doc_check.py", shell=True)


class LicenseCheckCommand(Command):
    description = "License Check"
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        subprocess.call(
            'python3 ./scripts/apache_license_check.py vwo/ tests/ setup.py --copyright "2019-2021 Wingify Software Pvt. Ltd."',
            shell=True,
        )


class ReleasePatchCommand(Command):
    description = "Patch Release"
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        subprocess.call("bumpversion patch", shell=True)


class ReleaseMinorCommand(Command):
    description = "Minor Release"
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        subprocess.call("bumpversion minor", shell=True)


class ReleaseMajorCommand(Command):
    description = "Major Release"
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        subprocess.call("bumpversion major", shell=True)


class PostDevelopCommand(develop):
    """Post-installation for development mode."""

    def run(self):
        print("\nRUNNING POST INSTALL DEVELOP SCRIPT \n")

        subprocess.call("pre-commit install;", shell=True)  # skipcq: BAN-B602
        # skipcq: BAN-B602
        subprocess.call("chmod +x post-install.sh; ./post-install.sh", shell=True)

        print("\nDONE: RUNNING POST INSTALL DEVELOP SCRIPT \n")

        develop.run(self)


setup(
    name="vwo-python-sdk",
    version="1.8.2",
    description="Python SDK for VWO server-side A/B Testing",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="VWO",
    author_email="dev@wingify.com",
    url="https://github.com/wingify/vwo-python-sdk",
    license="Apache License 2.0",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    cmdclass={
        "develop": PostDevelopCommand,
        "patch": ReleasePatchCommand,
        "minor": ReleaseMinorCommand,
        "major": ReleaseMajorCommand,
        "license_check": LicenseCheckCommand,
        "doc_check": DocCheckCommand,
    },
    packages=find_packages(exclude=["tests"]),
    install_requires=REQUIREMENTS,
)



# Copyright 2019-2021 Wingify Software Pvt. Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from ...enums.file_name_enum import FileNameEnum
from .operand_evaluator import OperandEvaluator
from ...enums.segments import OperandTypes, OperatorTypes
from ...helpers.generic_util import get_key_value

FILE = FileNameEnum.Services.SegmentEvaluator


class SegmentEvaluator:
    """ Class to evaluate segments defined in VWO app """

    def __init__(self):
        """ Initializes this class with VWOLogger and OperandEvaluator """
        self.operand_evaluator = OperandEvaluator()

    def evaluate(self, segments, custom_variables):
        """A parser which recursively evaluates the custom_variables passed against
        the expression tree represented by segments, and returns the result.

        Args:
            segments(dict): The segments representing the expression tree
            custom_variables(dict): Key/value pair of variables

        Returns:
            bool(result): True or False
        """

        operator, sub_segments = get_key_value(segments)
        if operator == OperatorTypes.NOT:
            return not self.evaluate(sub_segments, custom_variables)
        elif operator == OperatorTypes.AND:
            return all(self.evaluate(y, custom_variables) for y in sub_segments)
        elif operator == OperatorTypes.OR:
            return any(self.evaluate(y, custom_variables) for y in sub_segments)
        elif operator == OperandTypes.CUSTOM_VARIABLE:
            return self.operand_evaluator.evaluate_custom_variable(sub_segments, custom_variables)
        elif operator == OperandTypes.USER:
            return self.operand_evaluator.evaluate_user(sub_segments, custom_variables)



# Copyright 2019-2021 Wingify Software Pvt. Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..helpers import impression_util
from ..constants import constants
from ..constants.constants import API_METHODS
from ..helpers import campaign_util, validate_util
from ..enums.log_message_enum import LogMessageEnum
from ..enums.file_name_enum import FileNameEnum
from ..enums.log_level_enum import LogLevelEnum

FILE = FileNameEnum.Api.Activate


def _activate(vwo_instance, campaign_key, user_id, **kwargs):
    """ This API method: Gets the variation assigned for the user
        for the campaign and send the metrics to VWO server

    1. Validates the arguments being passed
    2. Checks if user is eligible to get bucketed into the campaign,
    3. Assigns the determinitic variation to the user(based on userId),
        if user becomes part of campaign
        If UserStorage is used, it will look into it for the
        variation and if found, no further processing is done
    4. Sends an impression call to VWO server to track user

    Args:
        campaign_key (string): unique campaign key
        user_id (string): ID assigned to a user

    Keywork Args:
        custom_variables (dict): Custom variables required for segmentation
        variation_targeting_variables (dict): Whitelisting variables to target users

    Returns:
        strings|None: If variation is assigned then variation-name
        otherwise None in case of user not becoming part
    """

    vwo_instance.logger.set_api(API_METHODS.ACTIVATE)
    # Retrieve custom variables
    custom_variables = kwargs.get("custom_variables")
    variation_targeting_variables = kwargs.get("variation_targeting_variables")

    # Validate input parameters
    if (
        not validate_util.is_valid_string(campaign_key)
        or not validate_util.is_valid_string(user_id)
        or (custom_variables is not None and not validate_util.is_valid_dict(custom_variables))
        or (
            variation_targeting_variables is not None and not validate_util.is_valid_dict(variation_targeting_variables)
        )
    ):  # noqa: E501
        vwo_instance.logger.log(
            LogLevelEnum.ERROR, LogMessageEnum.ERROR_MESSAGES.ACTIVATE_API_INVALID_PARAMS.format(file=FILE)
        )
        return None

    # Get the campaign settings
    campaign = campaign_util.get_campaign(vwo_instance.settings_file, campaign_key)

    # Validate campaign
    if not campaign:
        return None

    # Get campaign type
    campaign_type = campaign.get("type")

    # Validate valid api call
    if campaign_type != constants.CAMPAIGN_TYPES.VISUAL_AB:
        vwo_instance.logger.log(
            LogLevelEnum.ERROR,
            LogMessageEnum.ERROR_MESSAGES.INVALID_API.format(
                file=FILE, user_id=user_id, campaign_key=campaign_key, campaign_type=campaign_type
            ),
        )
        return None

    # Once the matching RUNNING campaign is found, assign the
    # deterministic variation to the user_id provided
    variation = vwo_instance.variation_decider.get_variation(
        user_id,
        campaign,
        custom_variables=custom_variables,
        variation_targeting_variables=variation_targeting_variables,
    )  # noqa: E501

    # Check if variation_name has been assigned
    if not variation:
        return None

    # Variation found, dispatch log to our servers
    impression = impression_util.create_impression(
        vwo_instance.settings_file, campaign.get("id"), variation.get("id"), user_id
    )
    vwo_instance.event_dispatcher.dispatch(impression)

    vwo_instance.logger.log(
        LogLevelEnum.INFO,
        LogMessageEnum.INFO_MESSAGES.MAIN_KEYS_FOR_IMPRESSION.format(
            file=FILE,
            campaign_id=impression.get("experiment_id"),
            user_id=impression.get("uId"),
            account_id=impression.get("account_id"),
            variation_id=impression.get("combination"),
        ),
    )

    return variation.get("name")



# Copyright 2019-2021 Wingify Software Pvt. Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Validate methods and parameters passed to the SDK """

import sys
import json
import jsonschema
from ..schemas.settings_file_schema import SETTINGS_FILE_SCHEMA
from ..constants.constants import LOG_LEVELS, GOAL_TYPES
from . import generic_util

services = {"logger": ["log"], "event_dispatcher": ["dispatch"], "user_storage": ["get", "set"]}


def is_valid_settings_file(settings_file):
    """ Validates the settings_file

    Args:
        settings_file (json):
            JSON object received from our server or somewhere else,
            must be json string representation.

    Returns:
        bool: Whether the settings_file is valid or not
    """
    try:
        settings_file = json.loads(settings_file)
    except Exception:
        return False
    try:
        jsonschema.validate(instance=settings_file, schema=SETTINGS_FILE_SCHEMA)
    except Exception:
        return False
    return True


def is_valid_service(service, service_name):
    """ Checks whether the service passed by the user contains the necessary methods or not

    Args:
        service (classobj): User defined class instance
        service_name (string): Name of the service

    Returns:
        bool: Whether the class instance provided is valid or not
    """
    service_attributes = services.get(service_name)
    if not service_attributes:
        return False
    for attr in service_attributes:
        if getattr(service, attr, None) is None:
            return False
    return True


def is_valid_log_level(level):
    """ Validates if the value passed is lies inside the vwo.enums.LogLevelEnum enum or not.

    Args:
        val (any type): value to be tested

    Returns:
        bool: True if valid LOG_LEVEL else False
    """
    return level in generic_util.get_attribute_values(LOG_LEVELS)


def is_valid_goal_type(goal_type):
    """ Validates if the value passed is lies inside the vwo.constants.GOAL_TYPE enum or not.

    Args:
        val (any type): value to be tested

    Returns:
        bool: True if valid GOAL_TYPE else False
    """
    return goal_type in generic_util.get_attribute_values(GOAL_TYPES)


def is_valid_dict(val):
    """ Validates if the value passed is of dict type or not.

    Args:
        val (any type): value to be tested

    Returns:
        bool: True if dict else False
    """
    return type(val) is dict


def is_valid_value(val):
    """ Validates if the value passed is a valid value,
    note: None is a valid value

    Args:
        val (any type): value to be tested

    Returns:
        bool: True if not None and bool(val) is True else False
    """
    return val is not None and bool(val)


def is_valid_non_zero_number(val):
    """ Validates if the value passed is a number & it is non zero.

    Args:
        val (any type): value to be tested

    Returns:
        bool: True if nonzero number else False
    """
    return type(val) == int and is_valid_value(val)


def is_valid_number(val):
    """ Validates if the value passed is a number(int/float) or not.

    Args:
        val (any type): value to be tested

    Returns:
        bool: True if number else False
    """
    return type(val) == int


def is_valid_unicode(val):
    """ Validates if the value passed is a python unicode value or not.

    Args:
        val (any type): value to be tested

    Returns:
        bool: True if unicode else False
    """
    if sys.version_info[0] < 3:
        return type(val) is unicode
    return False


def is_valid_string(val):
    """ Validates if the value passed is a valid python str or not.

    Args:
        val (any type): value to be tested

    Returns:
        bool: True if string else False
    """
    return (type(val) == str or is_valid_unicode(val)) and is_valid_value(val)


def is_valid_basic_data_type(val):
    """ Validates if the value passed is of basic data type or not.

    Args:
        val (any type): value to be tested

    Returns:
        bool: True if basic data type else False
    """
    return type(val) in [int, float, bool, str]



# Copyright 2019-2021 Wingify Software Pvt. Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Helper utility that creates an impression to send to server """

import json
from ..constants import constants
from ..helpers import generic_util, uuid_util, validate_util
from ..enums.log_message_enum import LogMessageEnum
from ..enums.file_name_enum import FileNameEnum
from ..enums.log_level_enum import LogLevelEnum
from ..logger import VWOLogger

try:
    from urllib import quote  # Python 2.X
except ImportError:
    from urllib.parse import quote  # Python 3+

FILE = FileNameEnum.Helpers.ImpressionUtil


def create_impression(settings_file, campaign_id, variation_id, user_id, goal_id=None, revenue=None):
    """ Creates the impression from the arguments passed according to
    call type

    Args:
        settings_file (dict): Settings file object
        campaign_id (string): Campaign identifier
        variation_id (string): Variation identifier
        user_id (string): User identifier
        goal_id (string|None): Goal identifier, if building track impression
        revenue (string|float|int|None):
            Number value, in any representation, if building track impression

    Returns:
        None|dict: None if campaign ID or variation ID is invalid,
            else impression(dict)
    """

    if not validate_util.is_valid_non_zero_number(campaign_id) or not validate_util.is_valid_string(user_id):
        return None

    is_track_user_api = True
    if goal_id is not None:
        is_track_user_api = False

    impression = get_common_properties(user_id, settings_file)

    impression.update(experiment_id=campaign_id, combination=variation_id)

    url = constants.HTTPS_PROTOCOL + constants.ENDPOINTS.BASE_URL
    logger = VWOLogger.getInstance()

    if is_track_user_api:
        impression.update(ed=json.dumps({"p": constants.PLATFORM}))
        impression.update(url=url + constants.ENDPOINTS.TRACK_USER)
        logger.log(
            LogLevelEnum.DEBUG,
            LogMessageEnum.DEBUG_MESSAGES.IMPRESSION_FOR_TRACK_USER.format(
                file=FILE, properties=json.dumps(impression)
            ),
        )
    else:
        impression.update(url=url + constants.ENDPOINTS.TRACK_GOAL)
        impression.update(goal_id=goal_id)
        if revenue:
            impression.update(r=revenue)
        logger.log(
            LogLevelEnum.DEBUG,
            LogMessageEnum.DEBUG_MESSAGES.IMPRESSION_FOR_TRACK_GOAL.format(
                file=FILE, properties=json.dumps(impression)
            ),
        )
    return impression


def get_common_properties(user_id, settings_file):
    """ Returns commonly used params for making requests to our servers.

    Args:
        user_id (string): Unique identification of user
        settings_file: settings file containing campaign data for extracting account_id

    Returns:
        properties(object): commonly used params for making call to our servers
    """

    account_id = settings_file.get("accountId")
    properties = {
        "random": generic_util.get_random_number(),
        "sdk": constants.SDK_NAME,
        "sdk-v": constants.SDK_VERSION,
        "ap": constants.PLATFORM,
        "sId": generic_util.get_current_unix_timestamp(),
        "u": uuid_util.generate_for(user_id, account_id),
        "account_id": account_id,
        "uId": quote(user_id.encode("utf-8")),
    }
    return properties
